import copy
import functools
import math
from dataclasses import dataclass

import cbor2

from vyper.codegen.ir_node import IRnode
from vyper.compiler.settings import OptimizationLevel
from vyper.evm.opcodes import get_opcodes, version_check
from vyper.exceptions import CodegenPanic, CompilerPanic
from vyper.ir.optimizer import COMMUTATIVE_OPS
from vyper.utils import MemoryPositions
from vyper.version import version_tuple

PUSH_OFFSET = 0x5F
DUP_OFFSET = 0x7F
SWAP_OFFSET = 0x8F


def num_to_bytearray(x):
    o = []
    while x > 0:
        o.insert(0, x % 256)
        x //= 256
    return o


def PUSH(x):
    bs = num_to_bytearray(x)
    # starting in shanghai, can do push0 directly with no immediates
    if len(bs) == 0 and not version_check(begin="shanghai"):
        bs = [0]
    return [f"PUSH{len(bs)}"] + bs


# push an exact number of bytes
def PUSH_N(x, n):
    o = []
    for _i in range(n):
        o.insert(0, x % 256)
        x //= 256
    assert x == 0
    return [f"PUSH{len(o)}"] + o


_next_symbol = 0


def mksymbol(name=""):
    global _next_symbol
    _next_symbol += 1

    return f"_sym_{name}{_next_symbol}"


def mkdebug(pc_debugger, ast_source):
    i = Instruction("DEBUG", ast_source)
    i.pc_debugger = pc_debugger
    return [i]


def is_symbol(i):
    return isinstance(i, str) and i.startswith("_sym_")


# basically something like a symbol which gets resolved
# during assembly, but requires 4 bytes of space.
# (should only happen in deploy code)
def is_mem_sym(i):
    return isinstance(i, str) and i.startswith("_mem_")


def is_ofst(sym):
    return isinstance(sym, str) and sym == "_OFST"


def _runtime_code_offsets(ctor_mem_size, runtime_codelen):
    # we need two numbers to calculate where the runtime code
    # should be copied to in memory (and making sure we don't
    # trample immutables, which are written to during the ctor
    # code): the memory allocated for the ctor and the length
    # of the runtime code.
    # after the ctor has run but before copying runtime code to
    # memory, the layout is
    # <ctor memory variables> ... | data section
    # and after copying runtime code to memory (immediately before
    # returning the runtime code):
    # <runtime code>          ... | data section
    # since the ctor memory variables and runtime code overlap,
    # we start allocating the data section from
    # `max(ctor_mem_size, runtime_code_size)`

    runtime_code_end = max(ctor_mem_size, runtime_codelen)
    runtime_code_start = runtime_code_end - runtime_codelen

    return runtime_code_start, runtime_code_end


# Calculate the size of PUSH instruction we need to handle all
# mem offsets in the code. For instance, if we only see mem symbols
# up to size 256, we can use PUSH1.
def calc_mem_ofst_size(ctor_mem_size):
    return math.ceil(math.log(ctor_mem_size + 1, 256))


# temporary optimization to handle stack items for return sequences
# like `return return_ofst return_len`. this is kind of brittle because
# it assumes the arguments are already on the stack, to be replaced
# by better liveness analysis.
# NOTE: modifies input in-place
def _rewrite_return_sequences(ir_node, label_params=None):
    args = ir_node.args

    if ir_node.value == "return":
        if args[0].value == "ret_ofst" and args[1].value == "ret_len":
            ir_node.args[0].value = "pass"
            ir_node.args[1].value = "pass"
    if ir_node.value == "exit_to":
        # handle exit from private function
        if args[0].value == "return_pc":
            ir_node.value = "jump"
            args[0].value = "pass"
        else:
            # handle jump to cleanup
            ir_node.value = "seq"

            _t = ["seq"]
            if "return_buffer" in label_params:
                _t.append(["pop", "pass"])

            dest = args[0].value
            # works for both internal and external exit_to
            more_args = ["pass" if t.value == "return_pc" else t for t in args[1:]]
            _t.append(["goto", dest] + more_args)
            ir_node.args = IRnode.from_list(_t, ast_source=ir_node.ast_source).args

    if ir_node.value == "label":
        label_params = set(t.value for t in ir_node.args[1].args)

    for t in args:
        _rewrite_return_sequences(t, label_params)


def _assert_false():
    global _revert_label
    # use a shared failure block for common case of assert(x).
    # in the future we might want to change the code
    # at _sym_revert0 to: INVALID
    return [_revert_label, "JUMPI"]


def _add_postambles(asm_ops):
    to_append = []

    global _revert_label

    _revert_string = [_revert_label, "JUMPDEST", *PUSH(0), "DUP1", "REVERT"]

    if _revert_label in asm_ops:
        # shared failure block
        to_append.extend(_revert_string)

    if len(to_append) > 0:
        # insert the postambles *before* runtime code
        # so the data section of the runtime code can't bork the postambles.
        runtime = None
        if isinstance(asm_ops[-1], list) and isinstance(asm_ops[-1][0], RuntimeHeader):
            runtime = asm_ops.pop()

        # for some reason there might not be a STOP at the end of asm_ops.
        # (generally vyper programs will have it but raw IR might not).
        asm_ops.append("STOP")
        asm_ops.extend(to_append)

        if runtime:
            asm_ops.append(runtime)

    # need to do this recursively since every sublist is basically
    # treated as its own program (there are no global labels.)
    for t in asm_ops:
        if isinstance(t, list):
            _add_postambles(t)


class Instruction(str):
    def __new__(cls, sstr, *args, **kwargs):
        return super().__new__(cls, sstr)

    def __init__(self, sstr, ast_source=None, error_msg=None):
        self.error_msg = error_msg
        self.pc_debugger = False

        self.ast_source = ast_source


def apply_line_numbers(func):
    @functools.wraps(func)
    def apply_line_no_wrapper(*args, **kwargs):
        code = args[0]
        ret = func(*args, **kwargs)

        new_ret = [
            Instruction(i, code.ast_source, code.error_msg)
            if isinstance(i, str) and not isinstance(i, Instruction)
            else i
            for i in ret
        ]
        return new_ret

    return apply_line_no_wrapper


@apply_line_numbers
def compile_to_assembly(code, optimize=OptimizationLevel.GAS):
    global _revert_label
    _revert_label = mksymbol("revert")

    # don't overwrite ir since the original might need to be output, e.g. `-f ir,asm`
    code = copy.deepcopy(code)
    _rewrite_return_sequences(code)

    res = _compile_to_assembly(code)

    _add_postambles(res)

    _relocate_segments(res)

    if optimize != OptimizationLevel.NONE:
        optimize_assembly(res)
    return res


# Compiles IR to assembly
@apply_line_numbers
def _compile_to_assembly(code, withargs=None, existing_labels=None, break_dest=None, height=0):
    if withargs is None:
        withargs = {}
    if not isinstance(withargs, dict):
        raise CompilerPanic(f"Incorrect type for withargs: {type(withargs)}")

    def _data_ofst_of(sym, ofst, height_):
        # e.g. _OFST _sym_foo 32
        assert is_symbol(sym) or is_mem_sym(sym)
        if isinstance(ofst.value, int):
            # resolve at compile time using magic _OFST op
            return ["_OFST", sym, ofst.value]
        else:
            # if we can't resolve at compile time, resolve at runtime
            ofst = _compile_to_assembly(ofst, withargs, existing_labels, break_dest, height_)
            return ofst + [sym, "ADD"]

    def _height_of(witharg):
        ret = height - withargs[witharg]
        if ret > 16:
            raise Exception("With statement too deep")
        return ret

    if existing_labels is None:
        existing_labels = set()
    if not isinstance(existing_labels, set):
        raise CompilerPanic(f"must be set(), but got {type(existing_labels)}")

    # Opcodes
    if isinstance(code.value, str) and code.value.upper() in get_opcodes():
        o = []
        for i, c in enumerate(code.args[::-1]):
            o.extend(_compile_to_assembly(c, withargs, existing_labels, break_dest, height + i))
        o.append(code.value.upper())
        return o

    # Numbers
    elif isinstance(code.value, int):
        if code.value < -(2**255):
            raise Exception(f"Value too low: {code.value}")
        elif code.value >= 2**256:
            raise Exception(f"Value too high: {code.value}")
        return PUSH(code.value % 2**256)

    # Variables connected to with statements
    elif isinstance(code.value, str) and code.value in withargs:
        return ["DUP" + str(_height_of(code.value))]

    # Setting variables connected to with statements
    elif code.value == "set":
        if len(code.args) != 2 or code.args[0].value not in withargs:
            raise Exception("Set expects two arguments, the first being a stack variable")
        if height - withargs[code.args[0].value] > 16:
            raise Exception("With statement too deep")
        return _compile_to_assembly(code.args[1], withargs, existing_labels, break_dest, height) + [
            "SWAP" + str(height - withargs[code.args[0].value]),
            "POP",
        ]

    # Pass statements
    # TODO remove "dummy"; no longer needed
    elif code.value in ("pass", "dummy"):
        return []

    # "mload" from data section of the currently executing code
    elif code.value == "dload":
        loc = code.args[0]

        o = []
        # codecopy 32 bytes to FREE_VAR_SPACE, then mload from FREE_VAR_SPACE
        o.extend(PUSH(32))
        o.extend(_data_ofst_of("_sym_code_end", loc, height + 1))
        o.extend(PUSH(MemoryPositions.FREE_VAR_SPACE) + ["CODECOPY"])
        o.extend(PUSH(MemoryPositions.FREE_VAR_SPACE) + ["MLOAD"])
        return o

    # batch copy from data section of the currently executing code to memory
    # (probably should have named this dcopy but oh well)
    elif code.value == "dloadbytes":
        dst = code.args[0]
        src = code.args[1]
        len_ = code.args[2]

        o = []
        o.extend(_compile_to_assembly(len_, withargs, existing_labels, break_dest, height))
        o.extend(_data_ofst_of("_sym_code_end", src, height + 1))
        o.extend(_compile_to_assembly(dst, withargs, existing_labels, break_dest, height + 2))
        o.extend(["CODECOPY"])
        return o

    # "mload" from the data section of (to-be-deployed) runtime code
    elif code.value == "iload":
        loc = code.args[0]

        o = []
        o.extend(_data_ofst_of("_mem_deploy_end", loc, height))
        o.append("MLOAD")

        return o

    # "mstore" to the data section of (to-be-deployed) runtime code
    elif code.value == "istore":
        loc = code.args[0]
        val = code.args[1]

        o = []
        o.extend(_compile_to_assembly(val, withargs, existing_labels, break_dest, height))
        o.extend(_data_ofst_of("_mem_deploy_end", loc, height + 1))
        o.append("MSTORE")

        return o

    # batch copy from memory to the data section of runtime code
    elif code.value == "istorebytes":
        raise Exception("unimplemented")

    # If statements (2 arguments, ie. if x: y)
    elif code.value == "if" and len(code.args) == 2:
        o = []
        o.extend(_compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height))
        end_symbol = mksymbol("join")
        o.extend(["ISZERO", end_symbol, "JUMPI"])
        o.extend(_compile_to_assembly(code.args[1], withargs, existing_labels, break_dest, height))
        o.extend([end_symbol, "JUMPDEST"])
        return o
    # If statements (3 arguments, ie. if x: y, else: z)
    elif code.value == "if" and len(code.args) == 3:
        o = []
        o.extend(_compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height))
        mid_symbol = mksymbol("else")
        end_symbol = mksymbol("join")
        o.extend(["ISZERO", mid_symbol, "JUMPI"])
        o.extend(_compile_to_assembly(code.args[1], withargs, existing_labels, break_dest, height))
        o.extend([end_symbol, "JUMP", mid_symbol, "JUMPDEST"])
        o.extend(_compile_to_assembly(code.args[2], withargs, existing_labels, break_dest, height))
        o.extend([end_symbol, "JUMPDEST"])
        return o

    # repeat(counter_location, start, rounds, rounds_bound, body)
    # basically a do-while loop:
    #
    # assert(rounds <= rounds_bound)
    # if (rounds > 0) {
    #   do {
    #     body;
    #   } while (++i != start + rounds)
    # }
    elif code.value == "repeat":
        o = []
        if len(code.args) != 5:  # pragma: nocover
            raise CompilerPanic("bad number of repeat args")

        i_name = code.args[0]
        start = code.args[1]
        rounds = code.args[2]
        rounds_bound = code.args[3]
        body = code.args[4]

        entry_dest, continue_dest, exit_dest = (
            mksymbol("loop_start"),
            mksymbol("loop_continue"),
            mksymbol("loop_exit"),
        )

        # stack: []
        o.extend(_compile_to_assembly(start, withargs, existing_labels, break_dest, height))

        o.extend(_compile_to_assembly(rounds, withargs, existing_labels, break_dest, height + 1))

        # stack: i

        # assert rounds <= round_bound
        if rounds != rounds_bound:
            # stack: i, rounds
            o.extend(
                _compile_to_assembly(
                    rounds_bound, withargs, existing_labels, break_dest, height + 2
                )
            )
            # stack: i, rounds, rounds_bound
            # assert 0 <= rounds <= rounds_bound (for rounds_bound < 2**255)
            # TODO this runtime assertion shouldn't fail for
            # internally generated repeats.
            o.extend(["DUP2", "GT"] + _assert_false())

            # stack: i, rounds
            # if (0 == rounds) { goto end_dest; }
            o.extend(["DUP1", "ISZERO", exit_dest, "JUMPI"])

        # stack: start, rounds
        if start.value != 0:
            o.extend(["DUP2", "ADD"])

        # stack: i, exit_i
        o.extend(["SWAP1"])

        if i_name.value in withargs:
            raise CompilerPanic(f"shadowed loop variable {i_name}")
        withargs[i_name.value] = height + 1

        # stack: exit_i, i
        o.extend([entry_dest, "JUMPDEST"])
        o.extend(
            _compile_to_assembly(
                body, withargs, existing_labels, (exit_dest, continue_dest, height + 2), height + 2
            )
        )

        del withargs[i_name.value]

        # clean up any stack items left by body
        o.extend(["POP"] * body.valency)

        # stack: exit_i, i
        # increment i:
        o.extend([continue_dest, "JUMPDEST", "PUSH1", 1, "ADD"])

        # stack: exit_i, i+1 (new_i)
        # if (exit_i != new_i) { goto entry_dest }
        o.extend(["DUP2", "DUP2", "XOR", entry_dest, "JUMPI"])
        o.extend([exit_dest, "JUMPDEST", "POP", "POP"])

        return o

    # Continue to the next iteration of the for loop
    elif code.value == "continue":
        if not break_dest:
            raise CompilerPanic("Invalid break")
        dest, continue_dest, break_height = break_dest
        return [continue_dest, "JUMP"]
    # Break from inside a for loop
    elif code.value == "break":
        if not break_dest:
            raise CompilerPanic("Invalid break")
        dest, continue_dest, break_height = break_dest

        n_local_vars = height - break_height
        # clean up any stack items declared in the loop body
        cleanup_local_vars = ["POP"] * n_local_vars
        return cleanup_local_vars + [dest, "JUMP"]
    # Break from inside one or more for loops prior to a return statement inside the loop
    elif code.value == "cleanup_repeat":
        if not break_dest:
            raise CompilerPanic("Invalid break")
        # clean up local vars and internal loop vars
        _, _, break_height = break_dest
        # except don't pop label params
        if "return_buffer" in withargs:
            break_height -= 1
        if "return_pc" in withargs:
            break_height -= 1
        return ["POP"] * break_height
    # With statements
    elif code.value == "with":
        o = []
        o.extend(_compile_to_assembly(code.args[1], withargs, existing_labels, break_dest, height))
        old = withargs.get(code.args[0].value, None)
        withargs[code.args[0].value] = height
        o.extend(
            _compile_to_assembly(code.args[2], withargs, existing_labels, break_dest, height + 1)
        )
        if code.args[2].valency:
            o.extend(["SWAP1", "POP"])
        else:
            o.extend(["POP"])
        if old is not None:
            withargs[code.args[0].value] = old
        else:
            del withargs[code.args[0].value]
        return o

    # runtime statement (used to deploy runtime code)
    elif code.value == "deploy":
        memsize = code.args[0].value  # used later to calculate _mem_deploy_start
        ir = code.args[1]
        immutables_len = code.args[2].value
        assert isinstance(memsize, int), "non-int memsize"
        assert isinstance(immutables_len, int), "non-int immutables_len"

        runtime_begin = mksymbol("runtime_begin")

        subcode = _compile_to_assembly(ir)

        o = []

        # COPY the code to memory for deploy
        o.extend(["_sym_subcode_size", runtime_begin, "_mem_deploy_start", "CODECOPY"])

        # calculate the len of runtime code
        o.extend(["_OFST", "_sym_subcode_size", immutables_len])  # stack: len
        o.extend(["_mem_deploy_start"])  # stack: len mem_ofst
        o.extend(["RETURN"])

        # since the asm data structures are very primitive, to make sure
        # assembly_to_evm is able to calculate data offsets correctly,
        # we pass the memsize via magic opcodes to the subcode
        subcode = [RuntimeHeader(runtime_begin, memsize, immutables_len)] + subcode

        # append the runtime code after the ctor code
        # `append(...)` call here is intentional.
        # each sublist is essentially its own program with its
        # own symbols.
        # in the later step when the "ir" block compiled to EVM,
        # symbols in subcode are resolved to position from start of
        # runtime-code (instead of position from start of bytecode).
        o.append(subcode)

        return o

    # Seq (used to piece together multiple statements)
    elif code.value == "seq":
        o = []
        for arg in code.args:
            o.extend(_compile_to_assembly(arg, withargs, existing_labels, break_dest, height))
            if arg.valency == 1 and arg != code.args[-1]:
                o.append("POP")
        return o
    # Seq without popping.
    # unreachable keyword produces INVALID opcode
    elif code.value == "assert_unreachable":
        o = _compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height)
        end_symbol = mksymbol("reachable")
        o.extend([end_symbol, "JUMPI", "INVALID", end_symbol, "JUMPDEST"])
        return o
    # Assert (if false, exit)
    elif code.value == "assert":
        o = _compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height)
        o.extend(["ISZERO"])
        o.extend(_assert_false())
        return o

    # SHA3 a single value
    elif code.value == "sha3_32":
        o = _compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height)
        o.extend(
            [
                *PUSH(MemoryPositions.FREE_VAR_SPACE),
                "MSTORE",
                *PUSH(32),
                *PUSH(MemoryPositions.FREE_VAR_SPACE),
                "SHA3",
            ]
        )
        return o
    # SHA3 a 64 byte value
    elif code.value == "sha3_64":
        o = _compile_to_assembly(code.args[0], withargs, existing_labels, break_dest, height)
        o.extend(_compile_to_assembly(code.args[1], withargs, existing_labels, break_dest, height))
        o.extend(
            [
                *PUSH(MemoryPositions.FREE_VAR_SPACE2),
                "MSTORE",
                *PUSH(MemoryPositions.FREE_VAR_SPACE),
                "MSTORE",
                *PUSH(64),
                *PUSH(MemoryPositions.FREE_VAR_SPACE),
                "SHA3",
            ]
        )
        return o
    elif code.value == "select":
        # b ^ ((a ^ b) * cond) where cond is 1 or 0
        # let t = a ^ b
        cond = code.args[0]
        a = code.args[1]
        b = code.args[2]

        o = []
        o.extend(_compile_to_assembly(b, withargs, existing_labels, break_dest, height))
        o.extend(_compile_to_assembly(a, withargs, existing_labels, break_dest, height + 1))
        # stack: b a
        o.extend(["DUP2", "XOR"])
        # stack: b t
        o.extend(_compile_to_assembly(cond, withargs, existing_labels, break_dest, height + 2))
        # stack: b t cond
        o.extend(["MUL", "XOR"])

        # stack: b ^ (t * cond)
        return o

    # <= operator
    elif code.value == "le":
        return _compile_to_assembly(
            IRnode.from_list(["iszero", ["gt", code.args[0], code.args[1]]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )
    # >= operator
    elif code.value == "ge":
        return _compile_to_assembly(
            IRnode.from_list(["iszero", ["lt", code.args[0], code.args[1]]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )
    # <= operator
    elif code.value == "sle":
        return _compile_to_assembly(
            IRnode.from_list(["iszero", ["sgt", code.args[0], code.args[1]]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )
    # >= operator
    elif code.value == "sge":
        return _compile_to_assembly(
            IRnode.from_list(["iszero", ["slt", code.args[0], code.args[1]]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )
    # != operator
    elif code.value == "ne":
        return _compile_to_assembly(
            IRnode.from_list(["iszero", ["eq", code.args[0], code.args[1]]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )

    # e.g. 95 -> 96, 96 -> 96, 97 -> 128
    elif code.value == "ceil32":
        # floor32(x) = x - x % 32 == x & 0b11..100000 == x & (~31)
        # ceil32(x) = floor32(x + 31) == (x + 31) & (~31)
        x = code.args[0]
        return _compile_to_assembly(
            IRnode.from_list(["and", ["add", x, 31], ["not", 31]]),
            withargs,
            existing_labels,
            break_dest,
            height,
        )

    elif code.value == "data":
        data_node = [DataHeader("_sym_" + code.args[0].value)]

        for c in code.args[1:]:
            if isinstance(c.value, int):
                assert 0 <= c < 256, f"invalid data byte {c}"
                data_node.append(c.value)
            elif isinstance(c.value, bytes):
                data_node.append(c.value)
            elif isinstance(c, IRnode):
                assert c.value == "symbol"
                data_node.extend(
                    _compile_to_assembly(c, withargs, existing_labels, break_dest, height)
                )
            else:
                raise ValueError(f"Invalid data: {type(c)} {c}")

        # intentionally return a sublist.
        return [data_node]

    # jump to a symbol, and push variable # of arguments onto stack
    elif code.value == "goto":
        o = []
        for i, c in enumerate(reversed(code.args[1:])):
            o.extend(_compile_to_assembly(c, withargs, existing_labels, break_dest, height + i))
        o.extend(["_sym_" + code.args[0].value, "JUMP"])
        return o
    elif code.value == "djump":
        o = []
        # "djump" compiles to a raw EVM jump instruction
        jump_target = code.args[0]
        o.extend(_compile_to_assembly(jump_target, withargs, existing_labels, break_dest, height))
        o.append("JUMP")
        return o
    # push a literal symbol
    elif code.value == "symbol":
        return ["_sym_" + code.args[0].value]
    # set a symbol as a location.
    elif code.value == "label":
        label_name = code.args[0].value
        assert isinstance(label_name, str)

        if label_name in existing_labels:
            raise Exception(f"Label with name {label_name} already exists!")
        else:
            existing_labels.add(label_name)

        if code.args[1].value != "var_list":
            raise CodegenPanic("2nd arg to label must be var_list")
        var_args = code.args[1].args

        body = code.args[2]

        # new scope
        height = 0
        withargs = {}

        for arg in reversed(var_args):
            assert isinstance(
                arg.value, str
            )  # already checked for higher up but only the paranoid survive
            withargs[arg.value] = height
            height += 1

        body_asm = _compile_to_assembly(
            body, withargs=withargs, existing_labels=existing_labels, height=height
        )
        # pop_scoped_vars = ["POP"] * height
        # for now, _rewrite_return_sequences forces
        # label params to be consumed implicitly
        pop_scoped_vars = []

        return ["_sym_" + label_name, "JUMPDEST"] + body_asm + pop_scoped_vars

    elif code.value == "unique_symbol":
        symbol = code.args[0].value
        assert isinstance(symbol, str)

        if symbol in existing_labels:
            raise Exception(f"symbol {symbol} already exists!")
        else:
            existing_labels.add(symbol)

        return []

    elif code.value == "exit_to":
        raise CodegenPanic("exit_to not implemented yet!")

    # inject debug opcode.
    elif code.value == "debugger":
        return mkdebug(pc_debugger=False, ast_source=code.ast_source)
    # inject debug opcode.
    elif code.value == "pc_debugger":
        return mkdebug(pc_debugger=True, ast_source=code.ast_source)
    else:  # pragma: no cover
        raise ValueError(f"Weird code element: {type(code)} {code}")


def getpos(node):
    return (node.lineno, node.col_offset, node.end_lineno, node.end_col_offset)


def note_line_num(line_number_map, pc, item):
    # Record AST attached to pc
    if isinstance(item, Instruction):
        if (ast_node := item.ast_source) is not None:
            ast_node = ast_node.get_original_node()
            if hasattr(ast_node, "node_id"):
                line_number_map["pc_raw_ast_map"][pc] = ast_node

        if item.error_msg is not None:
            line_number_map["error_map"][pc] = item.error_msg

    note_breakpoint(line_number_map, pc, item)


def note_breakpoint(line_number_map, pc, item):
    # Record line number attached to pc
    if item == "DEBUG":
        # Is PC debugger, create PC breakpoint.
        if item.pc_debugger:
            line_number_map["pc_breakpoints"].add(pc)
        # Create line number breakpoint.
        else:
            line_number_map["breakpoints"].add(item.lineno + 1)


_TERMINAL_OPS = ("JUMP", "RETURN", "REVERT", "STOP", "INVALID")


def _prune_unreachable_code(assembly):
    # delete code between terminal ops and JUMPDESTS as those are
    # unreachable
    changed = False
    i = 0
    while i < len(assembly) - 1:
        if assembly[i] in _TERMINAL_OPS:
            # find the next jumpdest or sublist
            for j in range(i + 1, len(assembly)):
                next_is_jumpdest = (
                    j < len(assembly) - 1
                    and is_symbol(assembly[j])
                    and assembly[j + 1] == "JUMPDEST"
                )
                next_is_list = isinstance(assembly[j], list)
                if next_is_jumpdest or next_is_list:
                    break
            else:
                # fixup an off-by-one if we made it to the end of the assembly
                # without finding an jumpdest or sublist
                j = len(assembly)
            changed = j > i + 1
            del assembly[i + 1 : j]

        i += 1

    return changed


def _prune_inefficient_jumps(assembly):
    # prune sequences `_sym_x JUMP _sym_x JUMPDEST` to `_sym_x JUMPDEST`
    changed = False
    i = 0
    while i < len(assembly) - 4:
        if (
            is_symbol(assembly[i])
            and assembly[i + 1] == "JUMP"
            and assembly[i] == assembly[i + 2]
            and assembly[i + 3] == "JUMPDEST"
        ):
            # delete _sym_x JUMP
            changed = True
            del assembly[i : i + 2]
        else:
            i += 1

    return changed


def _optimize_inefficient_jumps(assembly):
    # optimize sequences `_sym_common JUMPI _sym_x JUMP _sym_common JUMPDEST`
    # to `ISZERO _sym_x JUMPI _sym_common JUMPDEST`
    changed = False
    i = 0
    while i < len(assembly) - 6:
        if (
            is_symbol(assembly[i])
            and assembly[i + 1] == "JUMPI"
            and is_symbol(assembly[i + 2])
            and assembly[i + 3] == "JUMP"
            and assembly[i] == assembly[i + 4]
            and assembly[i + 5] == "JUMPDEST"
        ):
            changed = True
            assembly[i] = "ISZERO"
            assembly[i + 1] = assembly[i + 2]
            assembly[i + 2] = "JUMPI"
            del assembly[i + 3 : i + 4]
        else:
            i += 1

    return changed


def _merge_jumpdests(assembly):
    # When we have multiple JUMPDESTs in a row, or when a JUMPDEST
    # is immediately followed by another JUMP, we can skip the
    # intermediate jumps.
    # (Usually a chain of JUMPs is created by a nested block,
    # or some nested if statements.)
    changed = False
    i = 0
    while i < len(assembly) - 3:
        if is_symbol(assembly[i]) and assembly[i + 1] == "JUMPDEST":
            current_symbol = assembly[i]
            if is_symbol(assembly[i + 2]) and assembly[i + 3] == "JUMPDEST":
                # _sym_x JUMPDEST _sym_y JUMPDEST
                # replace all instances of _sym_x with _sym_y
                # (except for _sym_x JUMPDEST - don't want duplicate labels)
                new_symbol = assembly[i + 2]
                if new_symbol != current_symbol:
                    for j in range(len(assembly)):
                        if assembly[j] == current_symbol and i != j:
                            assembly[j] = new_symbol
                            changed = True
            elif is_symbol(assembly[i + 2]) and assembly[i + 3] == "JUMP":
                # _sym_x JUMPDEST _sym_y JUMP
                # replace all instances of _sym_x with _sym_y
                # (except for _sym_x JUMPDEST - don't want duplicate labels)
                new_symbol = assembly[i + 2]
                for j in range(len(assembly)):
                    if assembly[j] == current_symbol and i != j:
                        assembly[j] = new_symbol
                        changed = True

        i += 1

    return changed


_RETURNS_ZERO_OR_ONE = {
    "LT",
    "GT",
    "SLT",
    "SGT",
    "EQ",
    "ISZERO",
    "CALL",
    "STATICCALL",
    "CALLCODE",
    "DELEGATECALL",
}


def _merge_iszero(assembly):
    changed = False

    i = 0
    # list of opcodes that return 0 or 1
    while i < len(assembly) - 2:
        if (
            isinstance(assembly[i], str)
            and assembly[i] in _RETURNS_ZERO_OR_ONE
            and assembly[i + 1 : i + 3] == ["ISZERO", "ISZERO"]
        ):
            changed = True
            # drop the extra iszeros
            del assembly[i + 1 : i + 3]
        else:
            i += 1
    i = 0
    while i < len(assembly) - 3:
        # ISZERO ISZERO could map truthy to 1,
        # but it could also just be a no-op before JUMPI.
        if (
            assembly[i : i + 2] == ["ISZERO", "ISZERO"]
            and is_symbol(assembly[i + 2])
            and assembly[i + 3] == "JUMPI"
        ):
            changed = True
            del assembly[i : i + 2]
        else:
            i += 1

    return changed


# a symbol _sym_x in assembly can either mean to push _sym_x to the stack,
# or it can precede a location in code which we want to add to symbol map.
# this helper function tells us if we want to add the previous instruction
# to the symbol map.
def is_symbol_map_indicator(asm_node):
    return asm_node == "JUMPDEST"


def _prune_unused_jumpdests(assembly):
    changed = False

    used_jumpdests = set()

    # find all used jumpdests
    for i in range(len(assembly) - 1):
        if is_symbol(assembly[i]) and not is_symbol_map_indicator(assembly[i + 1]):
            used_jumpdests.add(assembly[i])

    for item in assembly:
        if isinstance(item, list) and isinstance(item[0], DataHeader):
            # add symbols used in data sections as they are likely
            # used for a jumptable.
            for t in item:
                if is_symbol(t):
                    used_jumpdests.add(t)

    # delete jumpdests that aren't used
    i = 0
    while i < len(assembly) - 2:
        if is_symbol(assembly[i]) and assembly[i] not in used_jumpdests:
            changed = True
            del assembly[i : i + 2]
        else:
            i += 1

    return changed


def _stack_peephole_opts(assembly):
    changed = False
    i = 0
    while i < len(assembly) - 2:
        if assembly[i : i + 3] == ["DUP1", "SWAP2", "SWAP1"]:
            changed = True
            del assembly[i + 2]
            assembly[i] = "SWAP1"
            assembly[i + 1] = "DUP2"
            continue
        # usually generated by with statements that return their input like
        # (with x (...x))
        if assembly[i : i + 3] == ["DUP1", "SWAP1", "POP"]:
            # DUP1 SWAP1 POP == no-op
            changed = True
            del assembly[i : i + 3]
            continue
        # usually generated by nested with statements that don't return like
        # (with x (with y ...))
        if assembly[i : i + 3] == ["SWAP1", "POP", "POP"]:
            # SWAP1 POP POP == POP POP
            changed = True
            del assembly[i]
            continue
        if (
            isinstance(assembly[i], str)
            and assembly[i].startswith("SWAP")
            and assembly[i] == assembly[i + 1]
        ):
            changed = True
            del assembly[i : i + 2]
        if assembly[i] == "SWAP1" and assembly[i + 1].lower() in COMMUTATIVE_OPS:
            changed = True
            del assembly[i]
        i += 1

    return changed


# optimize assembly, in place
def optimize_assembly(assembly):
    for x in assembly:
        if isinstance(x, list) and isinstance(x[0], RuntimeHeader):
            optimize_assembly(x)

    for _ in range(1024):
        changed = False

        changed |= _prune_unreachable_code(assembly)
        changed |= _merge_iszero(assembly)
        changed |= _merge_jumpdests(assembly)
        changed |= _prune_inefficient_jumps(assembly)
        changed |= _optimize_inefficient_jumps(assembly)
        changed |= _prune_unused_jumpdests(assembly)
        changed |= _stack_peephole_opts(assembly)

        if not changed:
            return

    raise CompilerPanic("infinite loop detected during assembly reduction")  # pragma: nocover


def adjust_pc_maps(pc_maps, ofst):
    assert ofst >= 0

    ret = {}
    # source breakpoints, don't need to modify
    ret["breakpoints"] = pc_maps["breakpoints"].copy()
    ret["pc_breakpoints"] = {pc + ofst for pc in pc_maps["pc_breakpoints"]}
    ret["pc_jump_map"] = {k + ofst: v for (k, v) in pc_maps["pc_jump_map"].items()}
    ret["pc_raw_ast_map"] = {k + ofst: v for (k, v) in pc_maps["pc_raw_ast_map"].items()}
    ret["error_map"] = {k + ofst: v for (k, v) in pc_maps["error_map"].items()}

    return ret


SYMBOL_SIZE = 2  # size of a PUSH instruction for a code symbol


def _data_to_evm(assembly, symbol_map):
    ret = bytearray()
    assert isinstance(assembly[0], DataHeader)
    for item in assembly[1:]:
        if is_symbol(item):
            symbol = symbol_map[item].to_bytes(SYMBOL_SIZE, "big")
            ret.extend(symbol)
        elif isinstance(item, int):
            ret.append(item)
        elif isinstance(item, bytes):
            ret.extend(item)
        else:
            raise ValueError(f"invalid data {type(item)} {item}")

    return ret


# predict what length of an assembly [data] node will be in bytecode
def _length_of_data(assembly):
    ret = 0
    assert isinstance(assembly[0], DataHeader)
    for item in assembly[1:]:
        if is_symbol(item):
            ret += SYMBOL_SIZE
        elif isinstance(item, int):
            assert 0 <= item < 256, f"invalid data byte {item}"
            ret += 1
        elif isinstance(item, bytes):
            ret += len(item)
        else:
            raise ValueError(f"invalid data {type(item)} {item}")

    return ret


@dataclass
class RuntimeHeader:
    label: str
    ctor_mem_size: int
    immutables_len: int

    def __repr__(self):
        return f"<RUNTIME {self.label} mem @{self.ctor_mem_size} imms @{self.immutables_len}>"


@dataclass
class DataHeader:
    label: str

    def __repr__(self):
        return f"DATA {self.label}"


def _relocate_segments(assembly):
    # relocate all data segments to the end, otherwise data could be
    # interpreted as PUSH instructions and mangle otherwise valid jumpdests
    # relocate all runtime segments to the end as well
    data_segments = []
    non_data_segments = []
    code_segments = []
    for t in assembly:
        if isinstance(t, list):
            if isinstance(t[0], DataHeader):
                data_segments.append(t)
            else:
                _relocate_segments(t)  # recurse
                assert isinstance(t[0], RuntimeHeader)
                code_segments.append(t)
        else:
            non_data_segments.append(t)
    assembly.clear()
    assembly.extend(non_data_segments)
    assembly.extend(code_segments)
    assembly.extend(data_segments)


# TODO: change API to split assembly_to_evm and assembly_to_source/symbol_maps
def assembly_to_evm(assembly, pc_ofst=0, insert_compiler_metadata=False):
    bytecode, source_maps, _ = assembly_to_evm_with_symbol_map(
        assembly, pc_ofst=pc_ofst, insert_compiler_metadata=insert_compiler_metadata
    )
    return bytecode, source_maps


def assembly_to_evm_with_symbol_map(assembly, pc_ofst=0, insert_compiler_metadata=False):
    """
    Assembles assembly into EVM

    assembly: list of asm instructions
    pc_ofst: when constructing the source map, the amount to offset all
             pcs by (no effect until we add deploy code source map)
    insert_compiler_metadata: whether to append vyper metadata to output
                            (should be true for runtime code)
    """
    line_number_map = {
        "breakpoints": set(),
        "pc_breakpoints": set(),
        "pc_jump_map": {0: "-"},
        "pc_raw_ast_map": {},
        "error_map": {},
    }

    pc = 0
    symbol_map = {}

    runtime_code, runtime_code_start, runtime_code_end = None, None, None

    # to optimize the size of deploy code - we want to use the smallest
    # PUSH instruction possible which can support all memory symbols
    # (and also works with linear pass symbol resolution)
    # to do this, we first do a single pass to compile any runtime code
    # and use that to calculate mem_ofst_size.
    mem_ofst_size, ctor_mem_size = None, None
    max_mem_ofst = 0
    for i, item in enumerate(assembly):
        if isinstance(item, list) and isinstance(item[0], RuntimeHeader):
            assert runtime_code is None, "Multiple subcodes"

            assert ctor_mem_size is None
            ctor_mem_size = item[0].ctor_mem_size

            runtime_code, runtime_map = assembly_to_evm(item[1:])

            runtime_code_start, runtime_code_end = _runtime_code_offsets(
                ctor_mem_size, len(runtime_code)
            )
            assert runtime_code_end - runtime_code_start == len(runtime_code)

        if is_ofst(item) and is_mem_sym(assembly[i + 1]):
            max_mem_ofst = max(assembly[i + 2], max_mem_ofst)

    if runtime_code_end is not None:
        mem_ofst_size = calc_mem_ofst_size(runtime_code_end + max_mem_ofst)

    data_section_lengths = []
    immutables_len = None

    # go through the code, resolving symbolic locations
    # (i.e. JUMPDEST locations) to actual code locations
    for i, item in enumerate(assembly):
        note_line_num(line_number_map, pc, item)
        if item == "DEBUG":
            continue  # skip debug

        # update pc_jump_map
        if item == "JUMP":
            last = assembly[i - 1]
            if is_symbol(last) and last.startswith("_sym_internal"):
                if last.endswith("cleanup"):
                    # exit an internal function
                    line_number_map["pc_jump_map"][pc] = "o"
                else:
                    # enter an internal function
                    line_number_map["pc_jump_map"][pc] = "i"
            else:
                # everything else
                line_number_map["pc_jump_map"][pc] = "-"
        elif item in ("JUMPI", "JUMPDEST"):
            line_number_map["pc_jump_map"][pc] = "-"

        # update pc
        if is_symbol(item):
            if is_symbol_map_indicator(assembly[i + 1]):
                # Don't increment pc as the symbol itself doesn't go into code
                if item in symbol_map:
                    raise CompilerPanic(f"duplicate jumpdest {item}")

                symbol_map[item] = pc
            else:
                pc += SYMBOL_SIZE + 1  # PUSH2 highbits lowbits
        elif is_mem_sym(item):
            # PUSH<n> item
            pc += mem_ofst_size + 1
        elif is_ofst(item):
            assert is_symbol(assembly[i + 1]) or is_mem_sym(assembly[i + 1])
            assert isinstance(assembly[i + 2], int)
            # [_OFST, _sym_foo, bar] -> PUSH2 (foo+bar)
            # [_OFST, _mem_foo, bar] -> PUSHN (foo+bar)
            pc -= 1
        elif isinstance(item, list) and isinstance(item[0], RuntimeHeader):
            # we are in initcode
            symbol_map[item[0].label] = pc
            # add source map for all items in the runtime map
            t = adjust_pc_maps(runtime_map, pc)
            for key in line_number_map:
                line_number_map[key].update(t[key])
            immutables_len = item[0].immutables_len
            pc += len(runtime_code)
            # grab lengths of data sections from the runtime
            for t in item:
                if isinstance(t, list) and isinstance(t[0], DataHeader):
                    data_section_lengths.append(_length_of_data(t))

        elif isinstance(item, list) and isinstance(item[0], DataHeader):
            symbol_map[item[0].label] = pc
            pc += _length_of_data(item)
        else:
            pc += 1

    bytecode_suffix = b""
    if insert_compiler_metadata:
        # this will hold true when we are in initcode
        assert immutables_len is not None
        metadata = (
            len(runtime_code),
            data_section_lengths,
            immutables_len,
            {"vyper": version_tuple},
        )
        bytecode_suffix += cbor2.dumps(metadata)
        # append the length of the footer, *including* the length
        # of the length bytes themselves.
        suffix_len = len(bytecode_suffix) + 2
        bytecode_suffix += suffix_len.to_bytes(2, "big")

    pc += len(bytecode_suffix)

    symbol_map["_sym_code_end"] = pc
    symbol_map["_mem_deploy_start"] = runtime_code_start
    symbol_map["_mem_deploy_end"] = runtime_code_end
    if runtime_code is not None:
        symbol_map["_sym_subcode_size"] = len(runtime_code)

    # TODO refactor into two functions, create symbol_map and assemble

    ret = bytearray()

    # now that all symbols have been resolved, generate bytecode
    # using the symbol map
    to_skip = 0
    for i, item in enumerate(assembly):
        if to_skip > 0:
            to_skip -= 1
            continue

        if item in ("DEBUG",):
            continue  # skippable opcodes

        elif is_symbol(item):
            # push a symbol to stack
            if not is_symbol_map_indicator(assembly[i + 1]):
                bytecode, _ = assembly_to_evm(PUSH_N(symbol_map[item], n=SYMBOL_SIZE))
                ret.extend(bytecode)

        elif is_mem_sym(item):
            bytecode, _ = assembly_to_evm(PUSH_N(symbol_map[item], n=mem_ofst_size))
            ret.extend(bytecode)

        elif is_ofst(item):
            # _OFST _sym_foo 32
            ofst = symbol_map[assembly[i + 1]] + assembly[i + 2]
            n = mem_ofst_size if is_mem_sym(assembly[i + 1]) else SYMBOL_SIZE
            bytecode, _ = assembly_to_evm(PUSH_N(ofst, n))
            ret.extend(bytecode)
            to_skip = 2

        elif isinstance(item, int):
            ret.append(item)
        elif isinstance(item, str) and item.upper() in get_opcodes():
            ret.append(get_opcodes()[item.upper()][0])
        elif item[:4] == "PUSH":
            ret.append(PUSH_OFFSET + int(item[4:]))
        elif item[:3] == "DUP":
            ret.append(DUP_OFFSET + int(item[3:]))
        elif item[:4] == "SWAP":
            ret.append(SWAP_OFFSET + int(item[4:]))
        elif isinstance(item, list) and isinstance(item[0], RuntimeHeader):
            ret.extend(runtime_code)
        elif isinstance(item, list) and isinstance(item[0], DataHeader):
            ret.extend(_data_to_evm(item, symbol_map))
        else:  # pragma: no cover
            # unreachable
            raise ValueError(f"Weird symbol in assembly: {type(item)} {item}")

    ret.extend(bytecode_suffix)

    line_number_map["breakpoints"] = list(line_number_map["breakpoints"])
    line_number_map["pc_breakpoints"] = list(line_number_map["pc_breakpoints"])
    return bytes(ret), line_number_map, symbol_map



import pytest

from vyper import compiler
from vyper.exceptions import StructureException, TypeMismatch

fail_list = [
    """
event Bar:
    _value: int128[4]

x: decimal[4]

@external
def foo():
    log Bar(self.x)
    """,
    """
event Bar:
    _value: int128[4]

@external
def foo():
    x: decimal[4] = [0.0, 0.0, 0.0, 0.0]
    log Bar(x)
    """,
    """
struct Foo:
    pass

@external
def foo():
    log Foo  # missing parens
    """,
    """
event Test:
    n: uint256

@external
def test():
    log Test(-7)
   """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_logging_fail(bad_code):
    with pytest.raises((TypeMismatch, StructureException)):
        compiler.compile_code(bad_code)


@pytest.mark.parametrize("mutability", ["@pure", "@view"])
@pytest.mark.parametrize("visibility", ["@internal", "@external"])
def test_logging_from_non_mutable(mutability, visibility):
    code = f"""
event Test:
    n: uint256

{visibility}
{mutability}
def test():
    log Test(1)
    """
    with pytest.raises(StructureException):
        compiler.compile_code(code)



from decimal import Decimal

import pytest
from eth.codecs import abi
from eth_utils import keccak

from tests.utils import ZERO_ADDRESS, decimal_to_int
from vyper.exceptions import StackTooDeep
from vyper.utils import int_bounds


def _make_tx(env, address, signature, values):
    # helper function to create data that will fail runtime clamp
    sig = keccak(signature.encode()).hex()[:8]
    data = "".join(int(i).to_bytes(32, "big", signed=i < 0).hex() for i in values)
    env.message_call(address, data=f"0x{sig}{data}")


def _make_abi_encode_tx(env, address, signature, input_types, values):
    # helper function to broadcast transactions where data is constructed from abi_encode
    sig = keccak(signature.encode()).hex()[:8]
    data = abi.encode(input_types, values).hex()
    env.message_call(address, data=f"0x{sig}{data}")


def _make_dynarray_data(offset, length, values):
    input = [offset] + [length] + values
    data = "".join(int(i).to_bytes(32, "big", signed=i < 0).hex() for i in input)
    return data


def _make_invalid_dynarray_tx(env, address, signature, data):
    sig = keccak(signature.encode()).hex()[:8]
    env.message_call(address, data=f"0x{sig}{data}")


def test_bytes_clamper(tx_failed, get_contract):
    clamper_test_code = """
@external
def foo(s: Bytes[3]) -> Bytes[3]:
    return s
    """

    c = get_contract(clamper_test_code)
    assert c.foo(b"ca") == b"ca"
    assert c.foo(b"cat") == b"cat"
    with tx_failed():
        c.foo(b"cate")


def test_bytes_clamper_multiple_slots(tx_failed, get_contract):
    clamper_test_code = """
@external
def foo(s: Bytes[40]) -> Bytes[40]:
    return s
    """

    data = b"this is exactly forty characters long!!!"
    c = get_contract(clamper_test_code)

    assert c.foo(data[:30]) == data[:30]
    assert c.foo(data) == data
    with tx_failed():
        c.foo(data + b"!")


def test_bytes_clamper_on_init(tx_failed, get_contract):
    clamper_test_code = """
foo: Bytes[3]

@deploy
def __init__(x: Bytes[3]):
    self.foo = x

@external
def get_foo() -> Bytes[3]:
    return self.foo
    """

    c = get_contract(clamper_test_code, b"cat")
    assert c.get_foo() == b"cat"

    with tx_failed():
        get_contract(clamper_test_code, b"cats")


@pytest.mark.parametrize("n", list(range(1, 33)))
def test_bytes_m_clamper_passing(env, get_contract, n):
    values = [b"\xff" * (i + 1) for i in range(n)]

    code = f"""
@external
def foo(s: bytes{n}) -> bytes{n}:
    return s
    """

    c = get_contract(code)
    for v in values:
        v = v.ljust(n, b"\x00")
        assert c.foo(v) == v


@pytest.mark.parametrize("n", list(range(1, 32)))  # bytes32 always passes
def test_bytes_m_clamper_failing(env, get_contract, tx_failed, n):
    values = []
    values.append(b"\x00" * n + b"\x80")  # just one bit set
    values.append(b"\xff" * n + b"\x80")  # n*8 + 1 bits set
    values.append(b"\x00" * 31 + b"\x01")  # bytes32
    values.append(b"\xff" * 32)  # bytes32
    values.append(bytes(range(32)))  # 0x00010203..1f
    values.append(bytes(range(1, 33)))  # 0x01020304..a0
    values.append(b"\xff" * 32)

    code = f"""
@external
def foo(s: bytes{n}) -> bytes{n}:
    return s
    """

    c = get_contract(code)
    for v in values:
        # munge for `_make_tx`
        int_value = int.from_bytes(v, byteorder="big")
        with tx_failed():
            _make_tx(env, c.address, f"foo(bytes{n})", [int_value])


@pytest.mark.parametrize("n", list(range(32)))
def test_sint_clamper_passing(env, get_contract, n):
    bits = 8 * (n + 1)
    lo, hi = int_bounds(True, bits)
    values = [-1, 0, 1, lo, hi]
    code = f"""
@external
def foo(s: int{bits}) -> int{bits}:
    return s
    """

    c = get_contract(code)
    for v in values:
        assert c.foo(v) == v


@pytest.mark.parametrize("n", list(range(31)))  # int256 does not clamp
def test_sint_clamper_failing(env, tx_failed, get_contract, n):
    bits = 8 * (n + 1)
    lo, hi = int_bounds(True, bits)
    values = [-(2**255), 2**255 - 1, lo - 1, hi + 1]
    code = f"""
@external
def foo(s: int{bits}) -> int{bits}:
    return s
    """

    c = get_contract(code)
    for v in values:
        with tx_failed():
            _make_tx(env, c.address, f"foo(int{bits})", [v])


@pytest.mark.parametrize("value", [True, False])
def test_bool_clamper_passing(env, get_contract, value):
    code = """
@external
def foo(s: bool) -> bool:
    return s
    """

    c = get_contract(code)
    assert c.foo(value) == value


@pytest.mark.parametrize("value", [2, 3, 4, 8, 16, 2**256 - 1])
def test_bool_clamper_failing(env, tx_failed, get_contract, value):
    code = """
@external
def foo(s: bool) -> bool:
    return s
    """

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(bool)", [value])


@pytest.mark.parametrize("value", [0] + [2**i for i in range(5)])
def test_flag_clamper_passing(env, get_contract, value):
    code = """
flag Roles:
    USER
    STAFF
    ADMIN
    MANAGER
    CEO

@external
def foo(s: Roles) -> Roles:
    return s
    """

    c = get_contract(code)
    assert c.foo(value) == value


@pytest.mark.parametrize("value", [2**i for i in range(5, 256)])
def test_flag_clamper_failing(env, tx_failed, get_contract, value):
    code = """
flag Roles:
    USER
    STAFF
    ADMIN
    MANAGER
    CEO

@external
def foo(s: Roles) -> Roles:
    return s
    """

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(uint256)", [value])


@pytest.mark.parametrize("n", list(range(32)))
def test_uint_clamper_passing(env, get_contract, n):
    bits = 8 * (n + 1)
    values = [0, 1, 2**bits - 1]
    code = f"""
@external
def foo(s: uint{bits}) -> uint{bits}:
    return s
    """

    c = get_contract(code)
    for v in values:
        assert c.foo(v) == v


@pytest.mark.parametrize("n", list(range(31)))  # uint256 has no failing cases
def test_uint_clamper_failing(env, tx_failed, get_contract, n):
    bits = 8 * (n + 1)
    values = [-1, -(2**255), 2**bits]
    code = f"""
@external
def foo(s: uint{bits}) -> uint{bits}:
    return s
    """
    c = get_contract(code)
    for v in values:
        with tx_failed():
            _make_tx(env, c.address, f"foo(uint{bits})", [v])


@pytest.mark.parametrize(
    "address",
    [
        ZERO_ADDRESS,
        "0x0000000000000000000000000000000000000001",
        "0xFFfFfFffFFfffFFfFFfFFFFFffFFFffffFfFFFfF",
    ],
)
def test_address_clamper_passing(get_contract, address):
    code = """
@external
def foo(s: address) -> address:
    return s
    """

    c = get_contract(code)
    assert c.foo(address) == address


@pytest.mark.parametrize("value", [2**160, 2**256 - 1])
def test_address_clamper_failing(env, tx_failed, get_contract, value):
    code = """
@external
def foo(s: address) -> address:
    return s
    """

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(address)", [value])


@pytest.mark.parametrize(
    "value",
    [
        0,
        1,
        -1,
        Decimal(2**167 - 1) / 10**10,
        -Decimal(2**167) / 10**10,
        "0.0",
        "1.0",
        "-1.0",
        "0.0000000001",
        "0.9999999999",
        "-0.0000000001",
        "-0.9999999999",
        "18707220957835557353007165858768422651595.9365500927",  # 2**167 - 1e-10
        "-18707220957835557353007165858768422651595.9365500928",  # -2**167
    ],
)
def test_decimal_clamper_passing(get_contract, value):
    code = """
@external
def foo(s: decimal) -> decimal:
    return s
    """

    c = get_contract(code)

    assert c.foo(decimal_to_int(value)) == decimal_to_int(value)


@pytest.mark.parametrize(
    "value",
    [
        2**167,
        -(2**167 + 1),
        187072209578355573530071658587684226515959365500928,  # 2 ** 167
        -187072209578355573530071658587684226515959365500929,  # - (2 ** 127 - 1e-10)
    ],
)
def test_decimal_clamper_failing(env, tx_failed, get_contract, value):
    code = """
@external
def foo(s: decimal) -> decimal:
    return s
    """

    c = get_contract(code)

    with tx_failed():
        _make_tx(env, c.address, "foo(int168)", [value])


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_int128_array_clamper_passing(get_contract, value):
    code = """
@external
def foo(a: uint256, b: int128[5], c: uint256) -> int128[5]:
    return b
    """

    # on both ends of the array we place a `uint256` that would fail the clamp check,
    # to ensure there are no off-by-one errors
    d = [value] * 5
    c = get_contract(code)
    assert c.foo(2**127, [value] * 5, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(5))
def test_int128_array_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    # ensure the invalid value is detected at all locations in the array
    code = """
@external
def foo(b: int128[5]) -> int128[5]:
    return b
    """

    values = [0] * 5
    values[idx] = bad_value

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(int128[5])", values)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_int128_array_looped_clamper_passing(get_contract, value):
    # when an array is > 5 items, the arg clamper runs in a loop to reduce bytecode size
    code = """
@external
def foo(a: uint256, b: int128[10], c: uint256) -> int128[10]:
    return b
    """

    d = [value] * 10
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(10))
def test_int128_array_looped_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    code = """
@external
def foo(b: int128[10]) -> int128[10]:
    return b
    """

    values = [0] * 10
    values[idx] = bad_value

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(int128[10])", values)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_multidimension_array_clamper_passing(get_contract, value):
    code = """
@external
def foo(a: uint256, b: int128[6][3][1][8], c: uint256) -> int128[6][3][1][8]:
    return b
    """

    # 6 * 3 * 1 * 8 = 144, the total number of values in our multidimensional array
    d = [[[[value] * 6] * 3] * 1] * 8
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(12))
def test_multidimension_array_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    code = """
@external
def foo(b: int128[6][1][2]) -> int128[6][1][2]:
    return b
    """

    values = [0] * 12
    values[idx] = bad_value

    c = get_contract(code)
    with tx_failed():
        _make_tx(env, c.address, "foo(int128[6][1][2]])", values)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_int128_dynarray_clamper_passing(get_contract, value):
    code = """
@external
def foo(a: uint256, b: DynArray[int128, 5], c: uint256) -> DynArray[int128, 5]:
    return b
    """

    # on both ends of the array we place a `uint256` that would fail the clamp check,
    # to ensure there are no off-by-one errors
    d = [value] * 5
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(5))
def test_int128_dynarray_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    # ensure the invalid value is detected at all locations in the array
    code = """
@external
def foo(b: int128[5]) -> int128[5]:
    return b
    """

    values = [0] * 5
    values[idx] = bad_value
    signature = "foo(int128[])"

    c = get_contract(code)

    data = _make_dynarray_data(32, 5, values)
    with tx_failed():
        _make_invalid_dynarray_tx(env, c.address, signature, data)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_int128_dynarray_looped_clamper_passing(get_contract, value):
    # when an array is > 5 items, the arg clamper runs in a loop to reduce bytecode size
    code = """
@external
def foo(a: uint256, b: DynArray[int128, 10], c: uint256) -> DynArray[int128, 10]:
    return b
    """
    d = [value] * 10
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(10))
def test_int128_dynarray_looped_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    code = """
@external
def foo(b: DynArray[int128, 10]) -> DynArray[int128, 10]:
    return b
    """

    values = [0] * 10
    values[idx] = bad_value

    c = get_contract(code)

    data = _make_dynarray_data(32, 10, values)
    signature = "foo(int128[])"
    with tx_failed():
        _make_invalid_dynarray_tx(env, c.address, signature, data)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
@pytest.mark.venom_xfail(raises=StackTooDeep, reason="stack scheduler regression")
def test_multidimension_dynarray_clamper_passing(get_contract, value):
    code = """
@external
def foo(
    a: uint256,
    b: DynArray[DynArray[DynArray[DynArray[int128, 5], 6], 7], 8],
    c: uint256
) -> DynArray[DynArray[DynArray[DynArray[int128, 5], 6], 7], 8]:
    return b
    """
    # Out of gas exception if outermost length is 6 and greater
    d = [[[[value] * 5] * 6] * 7] * 8
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(4))
def test_multidimension_dynarray_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    code = """
@external
def foo(b: DynArray[DynArray[int128, 2], 2]) -> DynArray[DynArray[int128, 2], 2]:
    return b
    """

    values = [[0] * 2] * 2
    values[idx // 2][idx % 2] = bad_value

    data = _make_dynarray_data(32, 2, [64, 160])  # Offset of nested arrays
    for v in values:
        # Length of 2
        v = [2] + v
        inner_data = "".join(int(_v).to_bytes(32, "big", signed=_v < 0).hex() for _v in v)
        data += inner_data

    signature = "foo(int128[][])"

    c = get_contract(code)
    with tx_failed():
        _make_invalid_dynarray_tx(env, c.address, signature, data)


@pytest.mark.parametrize("value", [0, 1, -1, 2**127 - 1, -(2**127)])
def test_dynarray_list_clamper_passing(get_contract, value):
    code = """
@external
def foo(
    a: uint256,
    b: DynArray[int128[5], 6],
    c: uint256
) -> DynArray[int128[5], 6]:
    return b
    """
    d = [[value] * 5] * 6
    c = get_contract(code)
    assert c.foo(2**127, d, 2**127) == d


@pytest.mark.parametrize("bad_value", [2**127, -(2**127) - 1, 2**255 - 1, -(2**255)])
@pytest.mark.parametrize("idx", range(10))
def test_dynarray_list_clamper_failing(env, tx_failed, get_contract, bad_value, idx):
    # ensure the invalid value is detected at all locations in the array
    code = """
@external
def foo(b: DynArray[int128[5], 2]) -> DynArray[int128[5], 2]:
    return b
    """

    values = [[0] * 5, [0] * 5]
    values[idx // 5][idx % 5] = bad_value

    data = _make_dynarray_data(32, 2, [])  # Offset of nested arrays
    for v in values:
        inner_data = "".join(int(_v).to_bytes(32, "big", signed=_v < 0).hex() for _v in v)
        data += inner_data

    c = get_contract(code)
    signature = "foo(int128[5][])"
    with tx_failed():
        _make_invalid_dynarray_tx(env, c.address, signature, data)



import pytest

from vyper import ast as vy_ast
from vyper.compiler import compile_code
from vyper.compiler.phases import CompilerData
from vyper.exceptions import NatSpecSyntaxException

test_code = """
'''
@title A simulator for Bug Bunny, the most famous Rabbit
@license MIT
@author Warned Bros
@notice You can use this contract for only the most basic simulation
@dev
    Simply chewing a carrot does not count, carrots must pass
    the throat to be considered eaten
'''

@external
@payable
def doesEat(food: String[30], qty: uint256) -> bool:
    '''
    @notice Determine if Bugs will accept `qty` of `food` to eat
    @dev Compares the entire string and does not rely on a hash
    @param food The name of a food to evaluate (in English)
    @param qty The number of food items to evaluate
    @return True if Bugs will eat it, False otherwise
    @custom:my-custom-tag hello, world!
    '''
    return True
"""


expected_userdoc = {
    "methods": {
        "doesEat(string,uint256)": {
            "notice": "Determine if Bugs will accept `qty` of `food` to eat"
        }
    },
    "notice": "You can use this contract for only the most basic simulation",
}


expected_devdoc = {
    "author": "Warned Bros",
    "license": "MIT",
    "details": "Simply chewing a carrot does not count, carrots must pass the throat to be considered eaten",  # NOQA: E501
    "methods": {
        "doesEat(string,uint256)": {
            "details": "Compares the entire string and does not rely on a hash",
            "params": {
                "food": "The name of a food to evaluate (in English)",
                "qty": "The number of food items to evaluate",
            },
            "returns": {"_0": "True if Bugs will eat it, False otherwise"},
            "custom:my-custom-tag": "hello, world!",
        }
    },
    "title": "A simulator for Bug Bunny, the most famous Rabbit",
}


def parse_natspec(code):
    vyper_ast = CompilerData(code).annotated_vyper_module
    return vy_ast.parse_natspec(vyper_ast)


def test_documentation_example_output():
    natspec = parse_natspec(test_code)

    assert natspec.userdoc == expected_userdoc
    assert natspec.devdoc == expected_devdoc


def test_no_tags_implies_notice():
    code = """
'''
Because there is no tag, this docstring is handled as a notice.
'''
@external
def foo():
    '''
    This one too!
    '''
    pass
    """

    natspec = parse_natspec(code)

    assert natspec.userdoc == {
        "methods": {"foo()": {"notice": "This one too!"}},
        "notice": "Because there is no tag, this docstring is handled as a notice.",
    }
    assert natspec.devdoc == {}


def test_whitespace():
    code = """
'''
        @dev

  Whitespace    gets  cleaned
    up,
            people can use


         awful formatting.


We don't mind!

@author Mr No-linter
                '''
"""
    natspec = parse_natspec(code)

    assert natspec.devdoc == {
        "author": "Mr No-linter",
        "details": "Whitespace gets cleaned up, people can use awful formatting. We don't mind!",
    }


def test_params():
    code = """
@external
def foo(bar: int128, baz: uint256, potato: bytes32):
    '''
    @param bar a number
    @param baz also a number
    @dev we didn't document potato, but that's ok
    '''
    pass
    """

    natspec = parse_natspec(code)

    assert natspec.devdoc == {
        "methods": {
            "foo(int128,uint256,bytes32)": {
                "details": "we didn't document potato, but that's ok",
                "params": {"bar": "a number", "baz": "also a number"},
            }
        }
    }


def test_returns():
    code = """
@external
def foo(bar: int128, baz: uint256) -> (int128, uint256):
    '''
    @return value of bar
    @return value of baz
    '''
    return bar, baz
    """

    natspec = parse_natspec(code)

    assert natspec.devdoc == {
        "methods": {
            "foo(int128,uint256)": {"returns": {"_0": "value of bar", "_1": "value of baz"}}
        }
    }


def test_ignore_private_methods():
    code = """
@external
def foo(bar: int128, baz: uint256):
    '''@dev I will be parsed.'''
    pass

@internal
def notfoo(bar: int128, baz: uint256):
    '''@dev I will not be parsed.'''
    pass
    """

    natspec = parse_natspec(code)

    assert natspec.devdoc["methods"] == {"foo(int128,uint256)": {"details": "I will be parsed."}}


def test_partial_natspec():
    code = """
@external
def foo():
    '''
    Regular comments preceding natspec is not allowed
    @notice this is natspec
    '''
    pass
    """

    with pytest.raises(
        NatSpecSyntaxException, match="NatSpec docstring opens with untagged comment"
    ):
        parse_natspec(code)


empty_field_cases = [
    """
    @notice
    @dev notice shouldn't be empty
    @author nobody
    """,
    """
    @dev notice shouldn't be empty
    @notice
    @author nobody
    """,
    """
    @dev notice shouldn't be empty
    @author nobody
    @notice
    """,
]


@pytest.mark.parametrize("bad_docstring", empty_field_cases)
def test_empty_field(bad_docstring):
    code = f"""
@external
def foo():
    '''{bad_docstring}'''
    pass
    """
    with pytest.raises(NatSpecSyntaxException, match="No description given for tag '@notice'"):
        parse_natspec(code)


def test_unknown_field():
    code = """
@external
def foo():
    '''
    @notice this is ok
    @thing this is bad
    '''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match="Unknown NatSpec field '@thing'"):
        parse_natspec(code)


@pytest.mark.parametrize("field", ["title", "license"])
def test_invalid_field(field):
    code = f"""
@external
def foo():
    '''@{field} function level docstrings cannot have titles'''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match=f"'@{field}' is not a valid field"):
        parse_natspec(code)


licenses = [
    "Apache-2.0",
    "Apache-2.0 OR MIT",
    "PSF-2.0 AND MIT",
    "Apache-2.0 AND (MIT OR GPL-2.0-only)",
]


@pytest.mark.parametrize("license", licenses)
def test_license(license):
    code = f"""
'''
@license {license}
'''
@external
def foo():
    pass
    """

    natspec = parse_natspec(code)

    assert natspec.devdoc == {"license": license}


fields = ["title", "author", "license", "notice", "dev"]


@pytest.mark.parametrize("field", fields)
def test_empty_fields(field):
    code = f"""
'''
@{field}
'''
@external
def foo():
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match=f"No description given for tag '@{field}'"):
        parse_natspec(code)


def test_duplicate_fields():
    code = """
@external
def foo():
    '''
    @notice It's fine to have one notice, but....
    @notice a second one, not so much
    '''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match="Duplicate NatSpec field '@notice'"):
        parse_natspec(code)


def test_duplicate_param():
    code = """
@external
def foo(bar: int128, baz: uint256):
    '''
    @param bar a number
    @param bar also a number
    '''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match="Parameter 'bar' documented more than once"):
        parse_natspec(code)


def test_unknown_param():
    code = """
@external
def foo(bar: int128, baz: uint256):
    '''@param hotdog not a number'''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match="Method has no parameter 'hotdog'"):
        parse_natspec(code)


empty_field_cases = [
    """
    @param a
    @dev param shouldn't be empty
    @author nobody
    """,
    """
    @dev param shouldn't be empty
    @param a
    @author nobody
    """,
    """
    @dev param shouldn't be empty
    @author nobody
    @param a
    """,
]


@pytest.mark.parametrize("bad_docstring", empty_field_cases)
def test_empty_param(bad_docstring):
    code = f"""
@external
def foo(a: int128):
    '''{bad_docstring}'''
    pass
    """
    with pytest.raises(NatSpecSyntaxException, match="No description given for parameter 'a'"):
        parse_natspec(code)


def test_too_many_returns_no_return_type():
    code = """
@external
def foo():
    '''@return should fail, the function does not include a return value'''
    pass
    """

    with pytest.raises(NatSpecSyntaxException, match="Method does not return any values"):
        parse_natspec(code)


def test_too_many_returns_single_return_type():
    code = """
@external
def foo() -> int128:
    '''
    @return int128
    @return this should fail
    '''
    return 1
    """

    with pytest.raises(
        NatSpecSyntaxException, match="Number of documented return values exceeds actual number"
    ):
        parse_natspec(code)


def test_too_many_returns_tuple_return_type():
    code = """
@external
def foo() -> (int128,uint256):
    '''
    @return int128
    @return uint256
    @return this should fail
    '''
    return 1, 2
    """

    with pytest.raises(
        NatSpecSyntaxException, match="Number of documented return values exceeds actual number"
    ):
        parse_natspec(code)


def test_natspec_parsed_implicitly():
    # test natspec is parsed even if not explicitly requested
    code = """
'''
@noticee x
'''
    """
    with pytest.raises(NatSpecSyntaxException):
        parse_natspec(code)

    # check we can get ast
    compile_code(code, output_formats=["ast_dict"])

    # anything beyond ast is blocked
    with pytest.raises(NatSpecSyntaxException):
        compile_code(code, output_formats=["annotated_ast_dict"])



import pytest

from vyper import compiler
from vyper.exceptions import VariableDeclarationException

fail_list = [
    """
q: int128 = 12
@external
def foo() -> int128:
    return self.q
    """,
    """
struct S:
    x: int128
s: S = S()
    """,
    """
foo.a: int128
    """,
    """
@external
def foo():
    bar.x: int128 = 0
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_variable_declaration_exception(bad_code):
    with pytest.raises(VariableDeclarationException):
        compiler.compile_code(bad_code)



import pytest
from eth.codecs.abi.exceptions import EncodeError

INITIAL_VALUE = 4


@pytest.fixture(scope="module")
def adv_storage_contract(get_contract):
    with open("examples/storage/advanced_storage.vy") as f:
        contract_code = f.read()
        # Pass constructor variables directly to the contract
        contract = get_contract(contract_code, INITIAL_VALUE)
    return contract


def test_initial_state(adv_storage_contract):
    # Check if the constructor of the contract is set up properly
    assert adv_storage_contract.storedData() == INITIAL_VALUE


def test_failed_transactions(env, adv_storage_contract, tx_failed):
    k1 = env.accounts[1]
    env.set_balance(k1, 10**18)

    # Try to set the storage to a negative amount
    with tx_failed():
        adv_storage_contract.set(-10, sender=k1)

    # Lock the contract by storing more than 100. Then try to change the value
    adv_storage_contract.set(150, sender=k1)
    with tx_failed():
        adv_storage_contract.set(10, sender=k1)

    # Reset the contract and try to change the value
    adv_storage_contract.reset(sender=k1)
    adv_storage_contract.set(10, sender=k1)
    assert adv_storage_contract.storedData() == 10

    # Assert a different exception (ValidationError for non-matching argument type)
    with tx_failed(EncodeError):
        adv_storage_contract.set("foo", sender=k1)

    # Assert a different exception that contains specific text
    with tx_failed(TypeError, "invocation failed due to improper number of arguments"):
        adv_storage_contract.set(1, 2, sender=k1)


def test_events(env, adv_storage_contract, get_logs):
    k1, k2 = env.accounts[:2]

    adv_storage_contract.set(10, sender=k1)
    (log1,) = get_logs(adv_storage_contract, "DataChange")
    adv_storage_contract.set(20, sender=k2)
    (log2,) = get_logs(adv_storage_contract, "DataChange")
    adv_storage_contract.reset(sender=k1)
    logs3 = get_logs(adv_storage_contract, "DataChange")

    # Check log contents
    assert log1.args.value == 10
    assert log2.args.setter == k2
    assert logs3 == []  # tx3 does not generate a log



import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold
from vyper.exceptions import OverflowException, TypeMismatch
from vyper.semantics.analysis.utils import validate_expected_type
from vyper.semantics.types.shortcuts import INT256_T, UINT256_T
from vyper.utils import unsigned_to_signed

st_uint256 = st.integers(min_value=0, max_value=2**256 - 1)

st_sint256 = st.integers(min_value=-(2**255), max_value=2**255 - 1)


# TODO: move this file to tests/unit/ast/nodes/test_fold_bitwise.py


@pytest.mark.fuzzing
@settings(max_examples=50)
@pytest.mark.parametrize("op", ["&", "|", "^"])
@given(a=st_uint256, b=st_uint256)
def test_bitwise_ops(get_contract, a, b, op):
    source = f"""
@external
def foo(a: uint256, b: uint256) -> uint256:
    return a {op} b
    """

    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{a} {op} {b}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(a, b) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=50)
@pytest.mark.parametrize("op", ["<<", ">>"])
@given(a=st_uint256, b=st.integers(min_value=0, max_value=256))
def test_bitwise_shift_unsigned(get_contract, a, b, op):
    source = f"""
@external
def foo(a: uint256, b: uint256) -> uint256:
    return a {op} b
    """
    contract = get_contract(source)

    try:
        vyper_ast = parse_and_fold(f"{a} {op} {b}")
        old_node = vyper_ast.body[0].value
        new_node = old_node.get_folded_value()
        # force bounds check, no-op because validate_numeric_bounds
        # already does this, but leave in for hygiene (in case
        # more types are added).
        validate_expected_type(new_node, UINT256_T)
    # compile time behavior does not match runtime behavior.
    # compile-time will throw on OOB, runtime will wrap.
    except OverflowException:  # here: check the wrapped value matches runtime
        assert op == "<<"
        assert contract.foo(a, b) == (a << b) % (2**256)
    else:
        assert contract.foo(a, b) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=51)
@pytest.mark.parametrize("op", ["<<", ">>"])
@given(a=st_sint256, b=st.integers(min_value=0, max_value=256))
@example(a=128, b=248)  # throws TypeMismatch
def test_bitwise_shift_signed(get_contract, a, b, op):
    source = f"""
@external
def foo(a: int256, b: uint256) -> int256:
    return a {op} b
    """
    contract = get_contract(source)

    try:
        vyper_ast = parse_and_fold(f"{a} {op} {b}")
        old_node = vyper_ast.body[0].value
        new_node = old_node.get_folded_value()
        validate_expected_type(new_node, INT256_T)  # force bounds check
    # compile time behavior does not match runtime behavior.
    # compile-time will throw on OOB, runtime will wrap.
    except (TypeMismatch, OverflowException):
        # check the wrapped value matches runtime
        assert op == "<<"
        assert contract.foo(a, b) == unsigned_to_signed((a << b) % (2**256), 256)
    else:
        assert contract.foo(a, b) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(value=st_uint256)
def test_bitwise_not(get_contract, value):
    source = """
@external
def foo(a: uint256) -> uint256:
    return ~a
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"~{value}")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(value) == new_node.value



def test_import_flag_types(make_input_bundle, get_contract):
    lib1 = """
import lib2

flag Roles:
    ADMIN
    USER

enum Roles2:
    ADMIN
    USER

role: Roles
role2: Roles2
role3: lib2.Roles3
    """
    lib2 = """
flag Roles3:
    ADMIN
    USER
    NOBODY
    """
    contract = """
import lib1

initializes: lib1

@external
def bar(r: lib1.Roles, r2: lib1.Roles2, r3: lib1.lib2.Roles3) -> bool:
    lib1.role = r
    lib1.role2 = r2
    lib1.role3 = r3
    assert lib1.role == lib1.Roles.ADMIN
    assert lib1.role2 == lib1.Roles2.USER
    assert lib1.role3 == lib1.lib2.Roles3.NOBODY
    return True
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})
    c = get_contract(contract, input_bundle=input_bundle)
    assert c.bar(1, 2, 4) is True



import pytest
from eth.codecs.abi.exceptions import EncodeError

TOKEN_NAME = "Vypercoin"
TOKEN_SYMBOL = "FANG"
TOKEN_DECIMALS = 18
TOKEN_INITIAL_SUPPLY = 21 * 10**6
TOKEN_TOTAL_SUPPLY = TOKEN_INITIAL_SUPPLY * (10**TOKEN_DECIMALS)


@pytest.fixture
def erc20(get_contract):
    with open("examples/tokens/ERC20.vy") as f:
        contract = get_contract(
            f.read(), *[TOKEN_NAME, TOKEN_SYMBOL, TOKEN_DECIMALS, TOKEN_INITIAL_SUPPLY]
        )
    return contract


@pytest.fixture
def erc20_caller(erc20, get_contract):
    erc20_caller_code = """
interface ERC20Contract:
    def name() -> String[64]: view
    def symbol() -> String[32]: view
    def decimals() -> uint256: view
    def balanceOf(_owner: address) -> uint256: view
    def totalSupply() -> uint256: view
    def transfer(_to: address, _amount: uint256) -> bool: nonpayable
    def transferFrom(_from: address, _to: address, _value: uint256) -> bool: nonpayable
    def approve(_spender: address, _amount: uint256) -> bool: nonpayable
    def allowance(_owner: address, _spender: address) -> uint256: nonpayable

token_address: ERC20Contract

@deploy
def __init__(token_addr: address):
    self.token_address = ERC20Contract(token_addr)

@external
def name() -> String[64]:
    return staticcall self.token_address.name()

@external
def symbol() -> String[32]:
    return staticcall self.token_address.symbol()

@external
def decimals() -> uint256:
    return staticcall self.token_address.decimals()

@external
def balanceOf(_owner: address) -> uint256:
    return staticcall self.token_address.balanceOf(_owner)

@external
def totalSupply() -> uint256:
    return staticcall self.token_address.totalSupply()

@external
def transfer(_to: address, _value: uint256) -> bool:
    return extcall self.token_address.transfer(_to, _value)

@external
def transferFrom(_from: address, _to: address, _value: uint256) -> bool:
    return extcall self.token_address.transferFrom(_from, _to, _value)

@external
def allowance(_owner: address, _spender: address) -> uint256:
    return extcall self.token_address.allowance(_owner, _spender)
    """
    return get_contract(erc20_caller_code, *[erc20.address])


def test_initial_state(env, erc20_caller):
    assert erc20_caller.totalSupply() == TOKEN_TOTAL_SUPPLY
    assert erc20_caller.balanceOf(env.deployer) == TOKEN_TOTAL_SUPPLY
    assert erc20_caller.balanceOf(env.accounts[1]) == 0
    assert erc20_caller.name() == TOKEN_NAME
    assert erc20_caller.symbol() == TOKEN_SYMBOL
    assert erc20_caller.decimals() == TOKEN_DECIMALS


def test_call_transfer(env, erc20, erc20_caller, tx_failed):
    # Basic transfer.
    erc20.transfer(erc20_caller.address, 10)
    assert erc20.balanceOf(erc20_caller.address) == 10
    erc20_caller.transfer(env.accounts[1], 10)
    assert erc20.balanceOf(erc20_caller.address) == 0
    assert erc20.balanceOf(env.accounts[1]) == 10

    # more than allowed
    with tx_failed():
        erc20_caller.transfer(env.accounts[1], TOKEN_TOTAL_SUPPLY)

    # Negative transfer value.
    with tx_failed(EncodeError):
        erc20_caller.transfer(env.accounts[1], -1)


def test_caller_approve_allowance(env, erc20, erc20_caller):
    assert erc20_caller.allowance(erc20.address, erc20_caller.address) == 0
    assert erc20.approve(erc20_caller.address, 10)
    assert erc20_caller.allowance(env.deployer, erc20_caller.address) == 10


def test_caller_tranfer_from(env, erc20, erc20_caller, tx_failed):
    # Cannot transfer tokens that are unavailable
    with tx_failed():
        erc20_caller.transferFrom(env.deployer, erc20_caller.address, 10)
    assert erc20.balanceOf(erc20_caller.address) == 0
    assert erc20.approve(erc20_caller.address, 10)
    erc20_caller.transferFrom(env.deployer, erc20_caller.address, 5)
    assert erc20.balanceOf(erc20_caller.address) == 5
    assert erc20_caller.allowance(env.deployer, erc20_caller.address) == 5
    erc20_caller.transferFrom(env.deployer, erc20_caller.address, 3)
    assert erc20.balanceOf(erc20_caller.address) == 8
    assert erc20_caller.allowance(env.deployer, erc20_caller.address) == 2



from functools import cached_property
from typing import TYPE_CHECKING, Optional

from vyper import ast as vy_ast
from vyper.abi_types import ABI_Address, ABIType
from vyper.ast.validation import validate_call_args
from vyper.exceptions import (
    InterfaceViolation,
    NamespaceCollision,
    StructureException,
    UnfoldableNode,
)
from vyper.semantics.analysis.base import Modifiability
from vyper.semantics.analysis.utils import (
    check_modifiability,
    get_exact_type_from_node,
    validate_expected_type,
    validate_unique_method_ids,
)
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types.base import TYPE_T, VyperType, is_type_t
from vyper.semantics.types.function import ContractFunctionT
from vyper.semantics.types.primitives import AddressT
from vyper.semantics.types.user import EventT, StructT, _UserType
from vyper.utils import OrderedSet, sha256sum

if TYPE_CHECKING:
    from vyper.semantics.analysis.base import ImportInfo, ModuleInfo


class InterfaceT(_UserType):
    typeclass = "interface"

    _type_members = {"address": AddressT()}
    _is_prim_word = True
    _as_array = True
    _as_hashmap_key = True
    _supports_external_calls = True
    _attribute_in_annotation = True

    def __init__(
        self,
        _id: str,
        decl_node: Optional[vy_ast.VyperNode],
        functions: dict,
        events: dict,
        structs: dict,
    ) -> None:
        validate_unique_method_ids(list(functions.values()))

        members = functions | events | structs

        # sanity check: by construction, there should be no duplicates.
        assert len(members) == len(functions) + len(events) + len(structs)

        super().__init__(functions)

        self._helper = VyperType(events | structs)
        self._id = _id
        self._helper._id = _id
        self.functions = functions
        self.events = events
        self.structs = structs

        self.decl_node = decl_node

    def get_type_member(self, attr, node):
        # get an event or struct from this interface
        return TYPE_T(self._helper.get_member(attr, node))

    @property
    def getter_signature(self):
        return (), AddressT()

    @property
    def abi_type(self) -> ABIType:
        return ABI_Address()

    def __repr__(self):
        return f"interface {self._id}"

    def _try_fold(self, node):
        if len(node.args) != 1:
            raise UnfoldableNode("wrong number of args", node.args)
        arg = node.args[0].get_folded_value()
        if not isinstance(arg, vy_ast.Hex):
            raise UnfoldableNode("not an address", arg)

        return node

    # when using the type itself (not an instance) in the call position
    def _ctor_call_return(self, node: vy_ast.Call) -> "InterfaceT":
        self._ctor_arg_types(node)
        return self

    def _ctor_arg_types(self, node):
        validate_call_args(node, 1)
        validate_expected_type(node.args[0], AddressT())
        return [AddressT()]

    def _ctor_kwarg_types(self, node):
        return {}

    def _ctor_modifiability_for_call(self, node: vy_ast.Call, modifiability: Modifiability) -> bool:
        return check_modifiability(node.args[0], modifiability)

    def validate_implements(
        self, node: vy_ast.ImplementsDecl, functions: dict[ContractFunctionT, vy_ast.VyperNode]
    ) -> None:
        fns_by_name = {fn_t.name: fn_t for fn_t in functions.keys()}

        unimplemented = []

        def _is_function_implemented(fn_name, fn_type):
            if fn_name not in fns_by_name:
                return False

            to_compare = fns_by_name[fn_name]
            assert isinstance(to_compare, ContractFunctionT)

            return to_compare.implements(fn_type)

        # check for missing functions
        for name, type_ in self.functions.items():
            if not isinstance(type_, ContractFunctionT):
                # ex. address
                continue

            if not _is_function_implemented(name, type_):
                unimplemented.append(type_._pp_signature)

        if len(unimplemented) > 0:
            # TODO: improve the error message for cases where the
            # mismatch is small (like mutability, or just one argument
            # is off, etc).
            missing_str = ", ".join(sorted(unimplemented))
            raise InterfaceViolation(
                f"Contract does not implement all interface functions: {missing_str}", node
            )

    def to_toplevel_abi_dict(self) -> list[dict]:
        abi = []
        for event in self.events.values():
            abi += event.to_toplevel_abi_dict()
        for func in self.functions.values():
            abi += func.to_toplevel_abi_dict()
        return abi

    # helper function which performs namespace collision checking
    @classmethod
    def _from_lists(
        cls,
        interface_name: str,
        decl_node: Optional[vy_ast.VyperNode],
        function_list: list[tuple[str, ContractFunctionT]],
        event_list: list[tuple[str, EventT]],
        struct_list: list[tuple[str, StructT]],
    ) -> "InterfaceT":
        functions = {}
        events = {}
        structs = {}

        seen_items: dict = {}

        def _mark_seen(name, item):
            if name in seen_items:
                msg = f"multiple functions or events named '{name}'!"
                prev_decl = seen_items[name].decl_node
                raise NamespaceCollision(msg, item.decl_node, prev_decl=prev_decl)
            seen_items[name] = item

        for name, function in function_list:
            _mark_seen(name, function)
            functions[name] = function

        for name, event in event_list:
            _mark_seen(name, event)
            events[name] = event

        for name, struct in struct_list:
            _mark_seen(name, struct)
            structs[name] = struct

        return cls(interface_name, decl_node, functions, events, structs)

    @classmethod
    def from_json_abi(cls, name: str, abi: dict) -> "InterfaceT":
        """
        Generate an `InterfaceT` object from an ABI.

        Arguments
        ---------
        name : str
            The name of the interface
        abi : dict
            Contract ABI

        Returns
        -------
        InterfaceT
            primitive interface type
        """
        functions: list = []
        events: list = []

        for item in [i for i in abi if i.get("type") == "function"]:
            functions.append((item["name"], ContractFunctionT.from_abi(item)))
        for item in [i for i in abi if i.get("type") == "event"]:
            events.append((item["name"], EventT.from_abi(item)))

        structs: list = []  # no structs in json ABI (as of yet)
        return cls._from_lists(name, None, functions, events, structs)

    @classmethod
    def from_ModuleT(cls, module_t: "ModuleT") -> "InterfaceT":
        """
        Generate an `InterfaceT` object from a Vyper ast node.

        Arguments
        ---------
        module_t: ModuleT
            Vyper module type
        Returns
        -------
        InterfaceT
            primitive interface type
        """
        funcs = []

        for fn_t in module_t.exposed_functions:
            funcs.append((fn_t.name, fn_t))

        if (fn_t := module_t.init_function) is not None:
            funcs.append((fn_t.name, fn_t))

        event_set: OrderedSet[EventT] = OrderedSet()
        event_set.update([node._metadata["event_type"] for node in module_t.event_defs])
        event_set.update(module_t.used_events)
        events = [(event_t.name, event_t) for event_t in event_set]

        # these are accessible via import, but they do not show up
        # in the ABI json
        structs = [(node.name, node._metadata["struct_type"]) for node in module_t.struct_defs]

        return cls._from_lists(module_t._id, module_t.decl_node, funcs, events, structs)

    @classmethod
    def from_InterfaceDef(cls, node: vy_ast.InterfaceDef) -> "InterfaceT":
        functions = []
        for func_ast in node.body:
            if not isinstance(func_ast, vy_ast.FunctionDef):
                raise StructureException(
                    "Interfaces can only contain function definitions", func_ast
                )
            if len(func_ast.decorator_list) > 0:
                raise StructureException(
                    "Function definition in interface cannot be decorated",
                    func_ast.decorator_list[0],
                )
            functions.append((func_ast.name, ContractFunctionT.from_InterfaceDef(func_ast)))

        # no structs or events in InterfaceDefs
        events: list = []
        structs: list = []

        return cls._from_lists(node.name, node, functions, events, structs)


# Datatype to store all module information.
class ModuleT(VyperType):
    typeclass = "module"

    _attribute_in_annotation = True
    _invalid_locations = (
        DataLocation.CALLDATA,
        DataLocation.CODE,
        DataLocation.MEMORY,
        DataLocation.TRANSIENT,
    )

    def __init__(self, module: vy_ast.Module, name: Optional[str] = None):
        super().__init__()

        self._module = module

        self._id = name or module.path

        # compute the interface, note this has the side effect of checking
        # for function collisions
        _ = self.interface

        self._helper = VyperType()
        self._helper._id = self._id

        for f in self.function_defs:
            # note: this checks for collisions
            self.add_member(f.name, f._metadata["func_type"])

        for item in self.exports_decls:
            for fn_t in item._metadata["exports_info"].functions:
                self.add_member(fn_t.name, fn_t)

        for e in self.event_defs:
            # add the type of the event so it can be used in call position
            self.add_member(e.name, TYPE_T(e._metadata["event_type"]))  # type: ignore

        for f in self.flag_defs:
            self.add_member(f.name, TYPE_T(f._metadata["flag_type"]))
            self._helper.add_member(f.name, TYPE_T(f._metadata["flag_type"]))

        for s in self.struct_defs:
            # add the type of the struct so it can be used in call position
            self.add_member(s.name, TYPE_T(s._metadata["struct_type"]))  # type: ignore
            self._helper.add_member(s.name, TYPE_T(s._metadata["struct_type"]))  # type: ignore

        for i in self.interface_defs:
            # add the type of the interface so it can be used in call position
            self.add_member(i.name, TYPE_T(i._metadata["interface_type"]))  # type: ignore

        for v in self.variable_decls:
            self.add_member(v.target.id, v.target._metadata["varinfo"])

        for i in self.import_stmts:
            import_info = i._metadata["import_info"]
            self.add_member(import_info.alias, import_info.typ)

            if hasattr(import_info.typ, "module_t"):
                self._helper.add_member(import_info.alias, TYPE_T(import_info.typ))

        for name, interface_t in self.interfaces.items():
            # can access interfaces in type position
            self._helper.add_member(name, TYPE_T(interface_t))

        self.add_member("__interface__", self.interface)

    # __eq__ is very strict on ModuleT - object equality! this is because we
    # don't want to reason about where a module came from (i.e. input bundle,
    # search path, symlinked vs normalized path, etc.)
    def __eq__(self, other):
        return self is other

    def __hash__(self):
        return hash(id(self))

    @property
    def decl_node(self) -> Optional[vy_ast.VyperNode]:  # type: ignore[override]
        return self._module

    def get_type_member(self, key: str, node: vy_ast.VyperNode) -> "VyperType":
        return self._helper.get_member(key, node)

    @cached_property
    def function_defs(self):
        return self._module.get_children(vy_ast.FunctionDef)

    @cached_property
    def event_defs(self):
        return self._module.get_children(vy_ast.EventDef)

    @cached_property
    def flag_defs(self):
        return self._module.get_children(vy_ast.FlagDef)

    @property
    def struct_defs(self):
        return self._module.get_children(vy_ast.StructDef)

    @property
    def interface_defs(self):
        return self._module.get_children(vy_ast.InterfaceDef)

    @cached_property
    def implements_decls(self):
        return self._module.get_children(vy_ast.ImplementsDecl)

    @cached_property
    def implemented_interfaces(self):
        ret = [node._metadata["interface_type"] for node in self.implements_decls]

        # a module implicitly implements module.__interface__.
        ret.append(self.interface)

        return ret

    @cached_property
    def interfaces(self) -> dict[str, InterfaceT]:
        ret = {}
        for i in self.interface_defs:
            assert i.name not in ret  # precondition
            ret[i.name] = i._metadata["interface_type"]

        for i in self.import_stmts:
            import_info = i._metadata["import_info"]
            if isinstance(import_info.typ, InterfaceT):
                assert import_info.alias not in ret  # precondition
                ret[import_info.alias] = import_info.typ

        return ret

    @property
    def import_stmts(self):
        return self._module.get_children((vy_ast.Import, vy_ast.ImportFrom))

    @cached_property
    def imported_modules(self) -> dict[str, "ModuleInfo"]:
        ret = {}
        for s in self.import_stmts:
            info = s._metadata["import_info"]
            module_info = info.typ
            if isinstance(module_info, InterfaceT):
                continue
            ret[info.alias] = module_info
        return ret

    @cached_property
    def reachable_imports(self) -> list["ImportInfo"]:
        """
        Return (recursively) reachable imports from this module as a list in
        depth-first (descendants-first) order.
        """
        ret = []
        for s in self.import_stmts:
            info = s._metadata["import_info"]

            # NOTE: this needs to be redone if interfaces can import other interfaces
            if not isinstance(info.typ, InterfaceT):
                ret.extend(info.typ.typ.reachable_imports)

            ret.append(info)

        return ret

    @cached_property
    def integrity_sum(self) -> str:
        acc = [sha256sum(self._module.full_source_code)]
        for s in self.import_stmts:
            info = s._metadata["import_info"]

            if isinstance(info.typ, InterfaceT):
                # NOTE: this needs to be redone if interfaces can import other interfaces
                acc.append(info.compiler_input.sha256sum)
            else:
                assert isinstance(info.typ.typ, ModuleT)
                acc.append(info.typ.typ.integrity_sum)

        return sha256sum("".join(acc))

    def find_module_info(self, needle: "ModuleT") -> Optional["ModuleInfo"]:
        for s in self.imported_modules.values():
            if s.module_t == needle:
                return s
        return None

    @property
    def variable_decls(self):
        return self._module.get_children(vy_ast.VariableDecl)

    @property
    def uses_decls(self):
        return self._module.get_children(vy_ast.UsesDecl)

    @property
    def initializes_decls(self):
        return self._module.get_children(vy_ast.InitializesDecl)

    @property
    def exports_decls(self):
        return self._module.get_children(vy_ast.ExportsDecl)

    @cached_property
    def used_modules(self):
        # modules which are written to
        ret = []
        for node in self.uses_decls:
            for used_module in node._metadata["uses_info"].used_modules:
                ret.append(used_module)
        return ret

    @property
    def initialized_modules(self):
        # modules which are initialized to
        ret = []
        for node in self.initializes_decls:
            info = node._metadata["initializes_info"]
            ret.append(info)
        return ret

    @cached_property
    def exposed_functions(self):
        # return external functions that are exposed in the runtime
        ret = []
        for node in self.exports_decls:
            ret.extend(node._metadata["exports_info"].functions)

        ret.extend([f for f in self.functions.values() if f.is_external])
        ret.extend([v.getter_ast._metadata["func_type"] for v in self.public_variables.values()])

        # precondition: no duplicate exports
        assert len(set(ret)) == len(ret)

        return ret

    @cached_property
    def init_function(self) -> Optional[ContractFunctionT]:
        return next((f for f in self.functions.values() if f.is_constructor), None)

    @cached_property
    def variables(self):
        # variables that this module defines, ex.
        # `x: uint256` is a private storage variable named x
        return {s.target.id: s.target._metadata["varinfo"] for s in self.variable_decls}

    @cached_property
    def public_variables(self):
        return {k: v for (k, v) in self.variables.items() if v.is_public}

    @cached_property
    def functions(self):
        return {f.name: f._metadata["func_type"] for f in self.function_defs}

    @cached_property
    # it would be nice to rely on the function analyzer to do this analysis,
    # but we don't have the result of function analysis at the time we need to
    # construct `self.interface`.
    def used_events(self) -> OrderedSet[EventT]:
        ret: OrderedSet[EventT] = OrderedSet()

        reachable: OrderedSet[ContractFunctionT] = OrderedSet()
        if self.init_function is not None:
            reachable.add(self.init_function)
            reachable.update(self.init_function.reachable_internal_functions)
        for fn_t in self.exposed_functions:
            reachable.add(fn_t)
            reachable.update(fn_t.reachable_internal_functions)

        for fn_t in reachable:
            fn_ast = fn_t.decl_node
            assert isinstance(fn_ast, vy_ast.FunctionDef)

            for node in fn_ast.get_descendants(vy_ast.Log):
                call_t = get_exact_type_from_node(node.value.func)
                if not is_type_t(call_t, EventT):
                    # this is an error, but it will be handled later
                    continue

                ret.add(call_t.typedef)

        return ret

    @cached_property
    def immutables(self):
        return [t for t in self.variables.values() if t.is_immutable]

    @cached_property
    def immutable_section_bytes(self):
        ret = 0
        for s in self.immutables:
            ret += s.typ.memory_bytes_required

        for initializes_info in self.initialized_modules:
            module_t = initializes_info.module_info.module_t
            ret += module_t.immutable_section_bytes

        return ret

    @cached_property
    def interface(self):
        return InterfaceT.from_ModuleT(self)



import pytest

from tests.utils import ZERO_ADDRESS

SOMEONE_TOKEN_IDS = [1, 2, 3]
OPERATOR_TOKEN_ID = 10
NEW_TOKEN_ID = 20
INVALID_TOKEN_ID = 99
ERC165_SIG = "0x01ffc9a7"
ERC165_INVALID_SIG = "0xffffffff"
ERC721_SIG = "0x80ac58cd"


@pytest.fixture(scope="module")
def c(get_contract, env):
    with open("examples/tokens/ERC721.vy") as f:
        code = f.read()
    c = get_contract(code)
    minter, someone, operator = env.accounts[:3]
    # someone owns 3 tokens
    for i in SOMEONE_TOKEN_IDS:
        c.mint(someone, i, sender=minter)
    # operator owns 1 tokens
    c.mint(operator, OPERATOR_TOKEN_ID, sender=minter)
    return c


def test_erc165(env, c):
    # From EIP-165:
    #   The source contract makes a STATICCALL to the destination address with input data:
    #       0x01ffc9a701ffc9a700000000000000000000000000000000000000000000000000000000
    #       and gas 30,000. This corresponds to `contract.supportsInterface(0x01ffc9a7)`
    assert c.supportsInterface(ERC165_SIG)
    #   If the call fails or return false, the destination contract does not implement ERC-165.
    #   If the call returns true, a second call is made with input data:
    #       0x01ffc9a7ffffffff00000000000000000000000000000000000000000000000000000000.
    assert not c.supportsInterface(ERC165_INVALID_SIG)
    #   If the second call fails or returns true, the destination contract does not implement
    #   ERC-165. Otherwise it implements ERC-165.

    assert c.supportsInterface(ERC721_SIG)


def test_balanceOf(c, env, tx_failed):
    someone = env.accounts[1]
    assert c.balanceOf(someone) == 3
    with tx_failed():
        c.balanceOf(ZERO_ADDRESS)


def test_ownerOf(c, env, tx_failed):
    someone = env.accounts[1]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[0]) == someone
    with tx_failed():
        c.ownerOf(INVALID_TOKEN_ID)


def test_getApproved(c, env):
    someone, operator = env.accounts[1:3]

    assert c.getApproved(SOMEONE_TOKEN_IDS[0]) == ZERO_ADDRESS

    c.approve(operator, SOMEONE_TOKEN_IDS[0], sender=someone)

    assert c.getApproved(SOMEONE_TOKEN_IDS[0]) == operator


def test_isApprovedForAll(c, env):
    someone, operator = env.accounts[1:3]

    assert c.isApprovedForAll(someone, operator) == 0

    c.setApprovalForAll(operator, True, sender=someone)

    assert c.isApprovedForAll(someone, operator) == 1


def test_transferFrom_by_owner(c, env, tx_failed, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer from zero address
    with tx_failed():
        c.transferFrom(ZERO_ADDRESS, operator, SOMEONE_TOKEN_IDS[0], sender=someone)

    # transfer to zero address
    with tx_failed():
        c.transferFrom(someone, ZERO_ADDRESS, SOMEONE_TOKEN_IDS[0], sender=someone)

    # transfer token without ownership
    with tx_failed():
        c.transferFrom(someone, operator, OPERATOR_TOKEN_ID, sender=someone)

    # transfer invalid token
    with tx_failed():
        c.transferFrom(someone, operator, INVALID_TOKEN_ID, sender=someone)

    # transfer by owner
    c.transferFrom(someone, operator, SOMEONE_TOKEN_IDS[0], sender=someone)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[0]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[0]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_transferFrom_by_approved(c, env, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer by approved
    c.approve(operator, SOMEONE_TOKEN_IDS[1], sender=someone)
    c.transferFrom(someone, operator, SOMEONE_TOKEN_IDS[1], sender=operator)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[1]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[1]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_transferFrom_by_operator(c, env, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer by operator
    c.setApprovalForAll(operator, True, sender=someone)
    c.transferFrom(someone, operator, SOMEONE_TOKEN_IDS[2], sender=operator)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[2]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[2]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_safeTransferFrom_by_owner(c, env, tx_failed, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer from zero address
    with tx_failed():
        c.safeTransferFrom(ZERO_ADDRESS, operator, SOMEONE_TOKEN_IDS[0], sender=someone)

    # transfer to zero address
    with tx_failed():
        c.safeTransferFrom(someone, ZERO_ADDRESS, SOMEONE_TOKEN_IDS[0], sender=someone)

    # transfer token without ownership
    with tx_failed():
        c.safeTransferFrom(someone, operator, OPERATOR_TOKEN_ID, sender=someone)

    # transfer invalid token
    with tx_failed():
        c.safeTransferFrom(someone, operator, INVALID_TOKEN_ID, sender=someone)

    # transfer by owner
    c.safeTransferFrom(someone, operator, SOMEONE_TOKEN_IDS[0], sender=someone)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[0]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[0]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_safeTransferFrom_by_approved(c, env, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer by approved
    c.approve(operator, SOMEONE_TOKEN_IDS[1], sender=someone)
    c.safeTransferFrom(someone, operator, SOMEONE_TOKEN_IDS[1], sender=operator)

    (log,) = get_logs(c, "Transfer")
    args = log.args
    assert args.sender == someone
    assert args.receiver == operator
    assert args.token_id == SOMEONE_TOKEN_IDS[1]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[1]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_safeTransferFrom_by_operator(c, env, get_logs):
    someone, operator = env.accounts[1:3]

    # transfer by operator
    c.setApprovalForAll(operator, True, sender=someone)
    c.safeTransferFrom(someone, operator, SOMEONE_TOKEN_IDS[2], sender=operator)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[2]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[2]) == operator
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(operator) == 2


def test_safeTransferFrom_to_contract(c, env, tx_failed, get_logs, get_contract):
    someone = env.accounts[1]

    # Can't transfer to a contract that doesn't implement the receiver code
    with tx_failed():
        c.safeTransferFrom(someone, c.address, SOMEONE_TOKEN_IDS[0], sender=someone)

    # Only to an address that implements that function
    receiver = get_contract(
        """
@external
def onERC721Received(
        _operator: address,
        _from: address,
        _tokenId: uint256,
        _data: Bytes[1024]
    ) -> bytes4:
    return method_id("onERC721Received(address,address,uint256,bytes)", output_type=bytes4)
    """
    )
    c.safeTransferFrom(someone, receiver.address, SOMEONE_TOKEN_IDS[0], sender=someone)

    (log,) = get_logs(c, "Transfer")
    assert log.args.sender == someone
    assert log.args.receiver == receiver.address
    assert log.args.token_id == SOMEONE_TOKEN_IDS[0]
    assert c.ownerOf(SOMEONE_TOKEN_IDS[0]) == receiver.address
    assert c.balanceOf(someone) == 2
    assert c.balanceOf(receiver.address) == 1


def test_approve(c, env, tx_failed, get_logs):
    someone, operator = env.accounts[1:3]

    # approve myself
    with tx_failed():
        c.approve(someone, SOMEONE_TOKEN_IDS[0], sender=someone)

    # approve token without ownership
    with tx_failed():
        c.approve(operator, OPERATOR_TOKEN_ID, sender=someone)

    # approve invalid token
    with tx_failed():
        c.approve(operator, INVALID_TOKEN_ID, sender=someone)

    c.approve(operator, SOMEONE_TOKEN_IDS[0], sender=someone)
    (log,) = get_logs(c, "Approval")

    assert log.args.owner == someone
    assert log.args.approved == operator
    assert log.args.token_id == SOMEONE_TOKEN_IDS[0]


def test_setApprovalForAll(c, env, tx_failed, get_logs):
    someone, operator = env.accounts[1:3]
    approved = True

    # setApprovalForAll myself
    with tx_failed():
        c.setApprovalForAll(someone, approved, sender=someone)

    c.setApprovalForAll(operator, approved, sender=someone)
    (log,) = get_logs(c, "ApprovalForAll")
    args = log.args
    assert args.owner == someone
    assert args.operator == operator
    assert args.approved == approved


def test_mint(c, env, tx_failed, get_logs):
    minter, someone = env.accounts[:2]

    # mint by non-minter
    with tx_failed():
        c.mint(someone, SOMEONE_TOKEN_IDS[0], sender=someone)

    # mint to zero address
    with tx_failed():
        c.mint(ZERO_ADDRESS, SOMEONE_TOKEN_IDS[0], sender=minter)

    # mint by minter
    c.mint(someone, NEW_TOKEN_ID, sender=minter)
    logs = get_logs(c, "Transfer")

    assert len(logs) > 0
    args = logs[0].args
    assert args.sender == ZERO_ADDRESS
    assert args.receiver == someone
    assert args.token_id == NEW_TOKEN_ID
    assert c.ownerOf(NEW_TOKEN_ID) == someone
    assert c.balanceOf(someone) == 4


def test_burn(c, env, tx_failed, get_logs):
    someone, operator = env.accounts[1:3]

    # burn token without ownership
    with tx_failed():
        c.burn(SOMEONE_TOKEN_IDS[0], sender=operator)

    # burn token by owner
    c.burn(SOMEONE_TOKEN_IDS[0], sender=someone)
    logs = get_logs(c, "Transfer")

    assert len(logs) > 0
    args = logs[0].args
    assert args.sender == someone
    assert args.receiver == ZERO_ADDRESS
    assert args.token_id == SOMEONE_TOKEN_IDS[0]
    with tx_failed():
        c.ownerOf(SOMEONE_TOKEN_IDS[0])
    assert c.balanceOf(someone) == 2



def test_uint256_mulmod(tx_failed, get_contract):
    uint256_code = """
@external
def _uint256_mulmod(x: uint256, y: uint256, z: uint256) -> uint256:
    return uint256_mulmod(x, y, z)
    """

    c = get_contract(uint256_code)

    assert c._uint256_mulmod(3, 1, 2) == 1
    assert c._uint256_mulmod(200, 3, 601) == 600
    assert c._uint256_mulmod(2**255, 1, 3) == 2
    assert c._uint256_mulmod(2**255, 2, 6) == 4
    with tx_failed():
        c._uint256_mulmod(2, 2, 0)


def test_uint256_mulmod_complex(get_contract):
    modexper = """
@external
def exponential(base: uint256, exponent: uint256, modulus: uint256) -> uint256:
    o: uint256 = 1
    for i: uint256 in range(256):
        o = uint256_mulmod(o, o, modulus)
        if exponent & shift(1, 255 - i) != 0:
            o = uint256_mulmod(o, base, modulus)
    return o
    """

    c = get_contract(modexper)
    assert c.exponential(3, 5, 100) == 43
    assert c.exponential(2, 997, 997) == 2


def test_uint256_mulmod_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: uint256) -> uint256: nonpayable

@external
def foo(f: Foo) -> uint256:
    return uint256_mulmod(200, 3, extcall f.foo(601))
    """

    c1 = side_effects_contract("uint256")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == 600

    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_uint256_mulmod_internal_call(get_contract):
    code = """
@external
def foo() -> uint256:
    return uint256_mulmod(self.a(), self.b(), self.c())

@internal
def a() -> uint256:
    return 200

@internal
def b() -> uint256:
    return 3

@internal
def c() -> uint256:
    return 601
    """

    c = get_contract(code)

    assert c.foo() == 600


def test_uint256_mulmod_evaluation_order(get_contract):
    code = """
a: uint256

@external
def foo1() -> uint256:
    self.a = 1
    return uint256_mulmod(self.a, 2, self.bar())

@external
def foo2() -> uint256:
    self.a = 1
    return uint256_mulmod(self.bar(), self.a, 2)

@external
def foo3() -> uint256:
    self.a = 1
    return uint256_mulmod(2, self.a, self.bar())

@internal
def bar() -> uint256:
    self.a = 7
    return 5
    """

    c = get_contract(code)

    assert c.foo1() == 2
    assert c.foo2() == 1
    assert c.foo3() == 2



import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from tests.utils import decimal_to_int, parse_and_fold
from vyper.utils import SizeLimits

st_decimals = st.decimals(
    min_value=SizeLimits.MIN_AST_DECIMAL,
    max_value=SizeLimits.MAX_AST_DECIMAL,
    allow_nan=False,
    allow_infinity=False,
    places=10,
)
st_int128 = st.integers(min_value=-(2**127), max_value=2**127 - 1)
st_uint256 = st.integers(min_value=0, max_value=2**256 - 1)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_decimals, right=st_decimals)
@pytest.mark.parametrize("fn_name", ["min", "max"])
def test_decimal(get_contract, left, right, fn_name):
    source = f"""
@external
def foo(a: decimal, b: decimal) -> decimal:
    return {fn_name}(a, b)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{fn_name}({left}, {right})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    l, r = [decimal_to_int(t) for t in (left, right)]
    assert contract.foo(l, r) == decimal_to_int(new_node.value)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_int128, right=st_int128)
@pytest.mark.parametrize("fn_name", ["min", "max"])
def test_int128(get_contract, left, right, fn_name):
    source = f"""
@external
def foo(a: int128, b: int128) -> int128:
    return {fn_name}(a, b)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{fn_name}({left}, {right})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(left, right) == new_node.value


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(left=st_uint256, right=st_uint256)
@pytest.mark.parametrize("fn_name", ["min", "max"])
def test_min_uint256(get_contract, left, right, fn_name):
    source = f"""
@external
def foo(a: uint256, b: uint256) -> uint256:
    return {fn_name}(a, b)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{fn_name}({left}, {right})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(left, right) == new_node.value



import pytest

from vyper.compiler import compile_code
from vyper.exceptions import CallViolation


def test_call_deploy_from_external(make_input_bundle):
    lib1 = """
@deploy
def __init__():
    pass
    """

    main = """
import lib1

@external
def foo():
    lib1.__init__()
    """

    input_bundle = make_input_bundle({"lib1.vy": lib1})

    with pytest.raises(CallViolation) as e:
        compile_code(main, input_bundle=input_bundle)

    assert e.value.message == "Cannot call an @deploy function from an @external function!"



# Author: Takayuki Jimba (@yudetamago), Ryuya Nakamura (@nrryuya)
# Modified from Philip Daian's tests:
# https://github.com/vyperlang/vyper/blob/v0.1.0-beta.5/tests/examples/tokens/ERC20_solidity_compatible/test/erc20_tests_1.py
import pytest
from eth.codecs.abi.exceptions import EncodeError

from tests.utils import ZERO_ADDRESS

MAX_UINT256 = (2**256) - 1  # Max uint256 value
TOKEN_NAME = "Vypercoin"
TOKEN_SYMBOL = "FANG"
TOKEN_DECIMALS = 18
TOKEN_INITIAL_SUPPLY = 0


@pytest.fixture(scope="module")
def c(get_contract):
    with open("examples/tokens/ERC20.vy") as f:
        code = f.read()
    return get_contract(code, *[TOKEN_NAME, TOKEN_SYMBOL, TOKEN_DECIMALS, TOKEN_INITIAL_SUPPLY])


@pytest.fixture(scope="module")
def c_bad(get_contract):
    # Bad contract is used for overflow checks on totalSupply corrupted
    with open("examples/tokens/ERC20.vy") as f:
        code = f.read()
    bad_code = code.replace("self.totalSupply += _value", "").replace(
        "self.totalSupply -= _value", ""
    )
    return get_contract(bad_code, *[TOKEN_NAME, TOKEN_SYMBOL, TOKEN_DECIMALS, TOKEN_INITIAL_SUPPLY])


def test_initial_state(c, env):
    a1, a2, a3 = env.accounts[1:4]
    # Check total supply, name, symbol and decimals are correctly set
    assert c.totalSupply() == TOKEN_INITIAL_SUPPLY
    assert c.name() == TOKEN_NAME
    assert c.symbol() == TOKEN_SYMBOL
    assert c.decimals() == TOKEN_DECIMALS
    # Check several account balances as 0
    assert c.balanceOf(a1) == 0
    assert c.balanceOf(a2) == 0
    assert c.balanceOf(a3) == 0
    # Check several allowances as 0
    assert c.allowance(a1, a1) == 0
    assert c.allowance(a1, a2) == 0
    assert c.allowance(a1, a3) == 0
    assert c.allowance(a2, a3) == 0


def test_mint_and_burn(c, env, tx_failed):
    minter, a1, a2 = env.accounts[0:3]

    # Test scenario were mints 2 to a1, burns twice (check balance consistency)
    assert c.balanceOf(a1) == 0
    c.mint(a1, 2, sender=minter)
    assert c.balanceOf(a1) == 2
    c.burn(2, sender=a1)
    assert c.balanceOf(a1) == 0
    with tx_failed():
        c.burn(2, sender=a1)
    assert c.balanceOf(a1) == 0
    # Test scenario were mintes 0 to a2, burns (check balance consistency, false burn)
    c.mint(a2, 0, sender=minter)
    assert c.balanceOf(a2) == 0
    with tx_failed():
        c.burn(2, sender=a2)
    # Check that a1 cannot burn after depleting their balance
    with tx_failed():
        c.burn(1, sender=a1)
    # Check that a1, a2 cannot mint
    with tx_failed():
        c.mint(a1, 1, sender=a1)
    with tx_failed():
        c.mint(a2, 1, sender=a2)
    # Check that mint to ZERO_ADDRESS failed
    with tx_failed():
        c.mint(ZERO_ADDRESS, 1, sender=a1)
    with tx_failed():
        c.mint(ZERO_ADDRESS, 1, sender=minter)


def test_totalSupply(c, env, tx_failed):
    # Test total supply initially, after mint, between two burns, and after failed burn
    minter, a1 = env.accounts[0:2]
    assert c.totalSupply() == 0
    c.mint(a1, 2, sender=minter)
    assert c.totalSupply() == 2
    c.burn(1, sender=a1)
    assert c.totalSupply() == 1
    c.burn(1, sender=a1)
    assert c.totalSupply() == 0
    with tx_failed():
        c.burn(1, sender=a1)
    assert c.totalSupply() == 0
    # Test that 0-valued mint can't affect supply
    c.mint(a1, 0, sender=minter)
    assert c.totalSupply() == 0


def test_transfer(c, env, tx_failed):
    minter, a1, a2 = env.accounts[0:3]
    with tx_failed():
        c.burn(1, sender=a2)
    c.mint(a1, 2, sender=minter)
    c.burn(1, sender=a1)
    c.transfer(a2, 1, sender=a1)
    with tx_failed():
        c.burn(1, sender=a1)
    c.burn(1, sender=a2)
    with tx_failed():
        c.burn(1, sender=a2)
    # Ensure transfer fails with insufficient balance
    with tx_failed():
        c.transfer(a1, 1, sender=a2)
    # Ensure 0-transfer always succeeds
    c.transfer(a1, 0, sender=a2)


def test_maxInts(c, env, tx_failed):
    minter, a1, a2 = env.accounts[0:3]
    c.mint(a1, MAX_UINT256, sender=minter)
    assert c.balanceOf(a1) == MAX_UINT256
    with tx_failed():
        c.mint(a1, 1, sender=a1)
    with tx_failed():
        c.mint(a1, MAX_UINT256, sender=a1)
    # Check that totalSupply cannot overflow, even when mint to other account
    with tx_failed():
        c.mint(a2, 1, sender=minter)
    # Check that corresponding mint is allowed after burn
    c.burn(1, sender=a1)
    c.mint(a2, 1, sender=minter)
    with tx_failed():
        c.mint(a2, 1, sender=minter)
    c.transfer(a1, 1, sender=a2)
    # Assert that after obtaining max number of tokens, a1 can transfer those but no more
    assert c.balanceOf(a1) == MAX_UINT256
    c.transfer(a2, MAX_UINT256, sender=a1)
    assert c.balanceOf(a2) == MAX_UINT256
    assert c.balanceOf(a1) == 0
    # [ next line should never work in EVM ]
    with pytest.raises(EncodeError):
        c.transfer(a1, MAX_UINT256 + 1, sender=a2)
    # Check approve/allowance w max possible token values
    assert c.balanceOf(a2) == MAX_UINT256
    c.approve(a1, MAX_UINT256, sender=a2)
    c.transferFrom(a2, a1, MAX_UINT256, sender=a1)
    assert c.balanceOf(a1) == MAX_UINT256
    assert c.balanceOf(a2) == 0
    # Check that max amount can be burned
    c.burn(MAX_UINT256, sender=a1)
    assert c.balanceOf(a1) == 0


def test_transferFrom_and_Allowance(c, env, tx_failed):
    minter, a1, a2, a3 = env.accounts[0:4]
    with tx_failed():
        c.burn(1, sender=a2)
    c.mint(a1, 1, sender=minter)
    c.mint(a2, 1, sender=minter)
    c.burn(1, sender=a1)
    # This should fail; no allowance or balance (0 always succeeds)
    with tx_failed():
        c.transferFrom(a1, a3, 1, sender=a2)
    c.transferFrom(a1, a3, 0, sender=a2)
    # Correct call to approval should update allowance (but not for reverse pair)
    c.approve(a2, 1, sender=a1)
    assert c.allowance(a1, a2) == 1
    assert c.allowance(a2, a1) == 0
    # transferFrom should succeed when allowed, fail with wrong sender
    with tx_failed():
        c.transferFrom(a1, a3, 1, sender=a3)
    assert c.balanceOf(a2) == 1
    c.approve(a1, 1, sender=a2)
    c.transferFrom(a2, a3, 1, sender=a1)
    # Allowance should be correctly updated after transferFrom
    assert c.allowance(a2, a1) == 0
    # transferFrom with no funds should fail despite approval
    c.approve(a1, 1, sender=a2)
    assert c.allowance(a2, a1) == 1
    with tx_failed():
        c.transferFrom(a2, a3, 1, sender=a1)
    # 0-approve should not change balance or allow transferFrom to change balance
    c.mint(a2, 1, sender=minter)
    assert c.allowance(a2, a1) == 1
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    with tx_failed():
        c.transferFrom(a2, a3, 1, sender=a1)
    # Test that if non-zero approval exists, 0-approval is NOT required to proceed
    # a non-conformant implementation is described in countermeasures at
    # https://docs.google.com/document/d/1YLPtQxZu1UAvO9cZ1O2RPXBbT0mooh4DYKjA_jp-RLM/edit#heading=h.m9fhqynw2xvt
    # the final spec insists on NOT using this behavior
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 1, sender=a2)
    assert c.allowance(a2, a1) == 1
    c.approve(a1, 2, sender=a2)
    assert c.allowance(a2, a1) == 2
    # Check that approving 0 then amount also works
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 5, sender=a2)
    assert c.allowance(a2, a1) == 5


def test_burnFrom_and_Allowance(c, env, tx_failed):
    minter, a1, a2, a3 = env.accounts[0:4]
    with tx_failed():
        c.burn(1, sender=a2)
    c.mint(a1, 1, sender=minter)
    c.mint(a2, 1, sender=minter)
    c.burn(1, sender=a1)
    # This should fail; no allowance or balance (0 always succeeds)
    with tx_failed():
        c.burnFrom(a1, 1, sender=a2)
    c.burnFrom(a1, 0, sender=a2)
    # Correct call to approval should update allowance (but not for reverse pair)
    c.approve(a2, 1, sender=a1)
    assert c.allowance(a1, a2) == 1
    assert c.allowance(a2, a1) == 0
    # transferFrom should succeed when allowed, fail with wrong sender
    with tx_failed():
        c.burnFrom(a2, 1, sender=a3)
    assert c.balanceOf(a2) == 1
    c.approve(a1, 1, sender=a2)
    c.burnFrom(a2, 1, sender=a1)
    # Allowance should be correctly updated after transferFrom
    assert c.allowance(a2, a1) == 0
    # transferFrom with no funds should fail despite approval
    c.approve(a1, 1, sender=a2)
    assert c.allowance(a2, a1) == 1
    with tx_failed():
        c.burnFrom(a2, 1, sender=a1)
    # 0-approve should not change balance or allow transferFrom to change balance
    c.mint(a2, 1, sender=minter)
    assert c.allowance(a2, a1) == 1
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    with tx_failed():
        c.burnFrom(a2, 1, sender=a1)
    # Test that if non-zero approval exists, 0-approval is NOT required to proceed
    # a non-conformant implementation is described in countermeasures at
    # https://docs.google.com/document/d/1YLPtQxZu1UAvO9cZ1O2RPXBbT0mooh4DYKjA_jp-RLM/edit#heading=h.m9fhqynw2xvt
    # the final spec insists on NOT using this behavior
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 1, sender=a2)
    assert c.allowance(a2, a1) == 1
    c.approve(a1, 2, sender=a2)
    assert c.allowance(a2, a1) == 2
    # Check that approving 0 then amount also works
    c.approve(a1, 0, sender=a2)
    assert c.allowance(a2, a1) == 0
    c.approve(a1, 5, sender=a2)
    assert c.allowance(a2, a1) == 5
    # Check that burnFrom to ZERO_ADDRESS failed
    with tx_failed():
        c.burnFrom(ZERO_ADDRESS, 0, sender=a1)


def test_raw_logs(c, env, get_logs):
    minter, a1, a2, a3 = env.accounts[0:4]

    def get_log_args(_, c, event_name):
        return get_logs(c, event_name)[0].args

    # Check that mint appropriately emits Transfer event
    args = get_log_args(c.mint(a1, 2, sender=minter), c, "Transfer")
    assert args.sender == ZERO_ADDRESS
    assert args.receiver == a1
    assert args.value == 2

    args = get_log_args(c.mint(a1, 0, sender=minter), c, "Transfer")
    assert args.sender == ZERO_ADDRESS
    assert args.receiver == a1
    assert args.value == 0

    # Check that burn appropriately emits Transfer event
    args = get_log_args(c.burn(1, sender=a1), c, "Transfer")
    assert args.sender == a1
    assert args.receiver == ZERO_ADDRESS
    assert args.value == 1

    args = get_log_args(c.burn(0, sender=a1), c, "Transfer")
    assert args.sender == a1
    assert args.receiver == ZERO_ADDRESS
    assert args.value == 0

    # Check that transfer appropriately emits Transfer event
    args = get_log_args(c.transfer(a2, 1, sender=a1), c, "Transfer")
    assert args.sender == a1
    assert args.receiver == a2
    assert args.value == 1

    args = get_log_args(c.transfer(a2, 0, sender=a1), c, "Transfer")
    assert args.sender == a1
    assert args.receiver == a2
    assert args.value == 0

    # Check that approving amount emits events
    args = get_log_args(c.approve(a1, 1, sender=a2), c, "Approval")
    assert args.owner == a2
    assert args.spender == a1
    assert args.value == 1

    args = get_log_args(c.approve(a2, 0, sender=a3), c, "Approval")
    assert args.owner == a3
    assert args.spender == a2
    assert args.value == 0

    # Check that transferFrom appropriately emits Transfer event
    args = get_log_args(c.transferFrom(a2, a3, 1, sender=a1), c, "Transfer")
    assert args.sender == a2
    assert args.receiver == a3
    assert args.value == 1

    args = get_log_args(c.transferFrom(a2, a3, 0, sender=a1), c, "Transfer")
    assert args.sender == a2
    assert args.receiver == a3
    assert args.value == 0


def test_bad_transfer(c_bad, env, tx_failed):
    # Ensure transfer fails if it would otherwise overflow balance when totalSupply is corrupted
    minter, a1, a2 = env.accounts[0:3]
    c_bad.mint(a1, MAX_UINT256, sender=minter)
    c_bad.mint(a2, 1, sender=minter)
    with tx_failed():
        c_bad.transfer(a1, 1, sender=a2)
    c_bad.transfer(a2, MAX_UINT256 - 1, sender=a1)
    assert c_bad.balanceOf(a1) == 1
    assert c_bad.balanceOf(a2) == MAX_UINT256


def test_bad_burn(c_bad, env, tx_failed):
    # Ensure burn fails if it would otherwise underflow balance when totalSupply is corrupted
    minter, a1 = env.accounts[0:2]
    assert c_bad.balanceOf(a1) == 0
    c_bad.mint(a1, 2, sender=minter)
    assert c_bad.balanceOf(a1) == 2
    with tx_failed():
        c_bad.burn(3, sender=a1)


def test_bad_transferFrom(c_bad, env, tx_failed):
    # Ensure transferFrom fails if it would otherwise overflow balance when totalSupply is corrupted
    minter, a1, a2 = env.accounts[0:3]
    c_bad.mint(a1, MAX_UINT256, sender=minter)
    c_bad.mint(a2, 1, sender=minter)
    c_bad.approve(a1, 1, sender=a2)
    with tx_failed():
        c_bad.transferFrom(a2, a1, 1, sender=a1)
    c_bad.approve(a2, MAX_UINT256 - 1, sender=a1)
    assert c_bad.allowance(a1, a2) == MAX_UINT256 - 1
    c_bad.transferFrom(a1, a2, MAX_UINT256 - 1, sender=a2)
    assert c_bad.balanceOf(a2) == MAX_UINT256



G1 = [1, 2]

G1_times_two = [
    1368015179489954701390400359078579693043519447331113978918064868415326638035,
    9918110051302171585080402603319702774565515993150576347155970296011118125764,
]

G1_times_three = [
    3353031288059533942658390886683067124040920775575537747144343083137631628272,
    19321533766552368860946552437480515441416830039777911637913418824951667761761,
]

negative_G1 = [1, 21888242871839275222246405745257275088696311157297823662689037894645226208581]

curve_order = 21888242871839275222246405745257275088548364400416034343698204186575808495617


def test_ecadd(get_contract):
    ecadder = """
x3: uint256[2]
y3: uint256[2]

@external
def _ecadd(x: uint256[2], y: uint256[2]) -> uint256[2]:
    return ecadd(x, y)

@external
def _ecadd2(x: uint256[2], y: uint256[2]) -> uint256[2]:
    x2: uint256[2] = x
    y2: uint256[2] = [y[0], y[1]]
    return ecadd(x2, y2)

@external
def _ecadd3(x: uint256[2], y: uint256[2]) -> uint256[2]:
    self.x3 = x
    self.y3 = [y[0], y[1]]
    return ecadd(self.x3, self.y3)

    """
    c = get_contract(ecadder)

    assert c._ecadd(G1, G1) == G1_times_two
    assert c._ecadd2(G1, G1_times_two) == G1_times_three
    assert c._ecadd3(G1, [0, 0]) == G1
    assert c._ecadd3(G1, negative_G1) == [0, 0]


def test_ecadd_internal_call(get_contract):
    code = """
@internal
def a() -> uint256[2]:
    return [1, 2]

@external
def foo() -> uint256[2]:
    return ecadd([1, 2], self.a())
    """
    c = get_contract(code)
    assert c.foo() == G1_times_two


def test_ecadd_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: uint256[2]) -> uint256[2]: payable

@external
def foo(a: Foo) -> uint256[2]:
    return ecadd([1, 2], extcall a.foo([1, 2]))
    """
    c1 = side_effects_contract("uint256[2]")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == G1_times_two

    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_ecadd_evaluation_order(get_contract):
    code = """
x: uint256[2]

@internal
def bar() -> uint256[2]:
    self.x = ecadd([1, 2], [1, 2])
    return [1, 2]

@external
def foo() -> bool:
    self.x = [1, 2]
    a: uint256[2] = ecadd([1, 2], [1, 2])
    b: uint256[2] = ecadd(self.x, self.bar())
    return a[0] == b[0] and a[1] == b[1]
    """
    c = get_contract(code)
    assert c.foo() is True


def test_ecmul(get_contract):
    ecmuller = """
x3: uint256[2]
y3: uint256

@external
def _ecmul(x: uint256[2], y: uint256) -> uint256[2]:
    return ecmul(x, y)

@external
def _ecmul2(x: uint256[2], y: uint256) -> uint256[2]:
    x2: uint256[2] = x
    y2: uint256 = y
    return ecmul(x2, y2)

@external
def _ecmul3(x: uint256[2], y: uint256) -> uint256[2]:
    self.x3 = x
    self.y3 = y
    return ecmul(self.x3, self.y3)

"""
    c = get_contract(ecmuller)

    assert c._ecmul(G1, 0) == [0, 0]
    assert c._ecmul(G1, 1) == G1
    assert c._ecmul(G1, 3) == G1_times_three
    assert c._ecmul(G1, curve_order - 1) == negative_G1
    assert c._ecmul(G1, curve_order) == [0, 0]


def test_ecmul_internal_call(get_contract):
    code = """
@internal
def a() -> uint256:
    return 3

@external
def foo() -> uint256[2]:
    return ecmul([1, 2], self.a())
    """
    c = get_contract(code)
    assert c.foo() == G1_times_three


def test_ecmul_ext_call(side_effects_contract, assert_side_effects_invoked, get_contract):
    code = """
interface Foo:
    def foo(x: uint256) -> uint256: payable

@external
def foo(a: Foo) -> uint256[2]:
    return ecmul([1, 2], extcall a.foo(3))
    """
    c1 = side_effects_contract("uint256")
    c2 = get_contract(code)

    assert c2.foo(c1.address) == G1_times_three

    assert_side_effects_invoked(c1, lambda: c2.foo(c1.address))


def test_ecmul_evaluation_order(get_contract):
    code = """
x: uint256[2]

@internal
def bar() -> uint256:
    self.x = ecmul([1, 2], 3)
    return 3

@external
def foo() -> bool:
    self.x = [1, 2]
    a: uint256[2] = ecmul([1, 2], 3)
    b: uint256[2] = ecmul(self.x, self.bar())
    return a[0] == b[0] and a[1] == b[1]
    """
    c = get_contract(code)
    assert c.foo() is True



import re

from vyper.exceptions import StructureException


def validate_identifier(attr, ast_node=None):
    if not re.match("^[_a-zA-Z][a-zA-Z0-9_]*$", attr):
        raise StructureException(f"'{attr}' contains invalid character(s)", ast_node)
    if attr.lower() in RESERVED_KEYWORDS:
        raise StructureException(f"'{attr}' is a reserved keyword", ast_node)


# https://docs.python.org/3/reference/lexical_analysis.html#keywords
# note we don't technically need to block all python reserved keywords,
# but do it for hygiene
_PYTHON_RESERVED_KEYWORDS = {
    "False",
    "None",
    "True",
    "and",
    "as",
    "assert",
    "async",
    "await",
    "break",
    "class",
    "continue",
    "def",
    "del",
    "elif",
    "else",
    "except",
    "finally",
    "for",
    "from",
    "global",
    "if",
    "import",
    "in",
    "is",
    "lambda",
    "nonlocal",
    "not",
    "or",
    "pass",
    "raise",
    "return",
    "try",
    "while",
    "with",
    "yield",
}
_PYTHON_RESERVED_KEYWORDS = {s.lower() for s in _PYTHON_RESERVED_KEYWORDS}

# Cannot be used for variable or member naming
RESERVED_KEYWORDS = _PYTHON_RESERVED_KEYWORDS | {
    # decorators
    "public",
    "external",
    "nonpayable",
    "constant",
    "immutable",
    "transient",
    "internal",
    "payable",
    "nonreentrant",
    # "class" keywords
    "interface",
    "struct",
    "event",
    "enum",
    "flag"
    # EVM operations
    "unreachable",
    # special functions (no name mangling)
    "init",
    "_init_",
    "___init___",
    "____init____",
    "default",
    "_default_",
    "___default___",
    "____default____",
    # more control flow and special operations
    "range",
    # more special operations
    "indexed",
    # denominations
    "ether",
    "wei",
    "finney",
    "szabo",
    "shannon",
    "lovelace",
    "ada",
    "babbage",
    "gwei",
    "kwei",
    "mwei",
    "twei",
    "pwei",
    # sentinel constant values
    # TODO remove when these are removed from the language
    "zero_address",
    "empty_bytes32",
    "max_int128",
    "min_int128",
    "max_decimal",
    "min_decimal",
    "max_uint256",
    "zero_wei",
}



from .external_function import generate_ir_for_external_function
from .internal_function import generate_ir_for_internal_function

__all__ = [generate_ir_for_internal_function, generate_ir_for_external_function]  # type: ignore



import pytest

from vyper.exceptions import UnknownType
from vyper.semantics.types import PRIMITIVE_TYPES, SArrayT
from vyper.semantics.types.utils import type_from_abi

BASE_TYPES = ["int128", "uint256", "bool", "address", "bytes32"]


@pytest.mark.parametrize("type_str", BASE_TYPES)
def test_base_types(type_str):
    base_t = PRIMITIVE_TYPES[type_str]
    type_t = type_from_abi({"type": type_str})

    assert base_t == type_t


@pytest.mark.parametrize("type_str", BASE_TYPES)
def test_base_types_as_arrays(type_str):
    base_t = PRIMITIVE_TYPES[type_str]
    type_t = type_from_abi({"type": f"{type_str}[3]"})

    assert type_t == SArrayT(base_t, 3)


@pytest.mark.parametrize("type_str", BASE_TYPES)
def test_base_types_as_multidimensional_arrays(type_str):
    base_t = PRIMITIVE_TYPES[type_str]

    type_t = type_from_abi({"type": f"{type_str}[3][5]"})

    assert type_t == SArrayT(SArrayT(base_t, 3), 5)


@pytest.mark.parametrize("idx", ["0", "-1", "0x00", "'1'", "foo", "[1]", "(1,)"])
def test_invalid_index(idx):
    with pytest.raises(UnknownType):
        type_from_abi({"type": f"int128[{idx}]"})
